// PIV: ANSI C implementation of the API in SP 800-73, Chapter 6 
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The PIV Middleware Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include "piv.h"

#include "tlvparse.h"

#include "pivCardEdge.h"
#include "pivtlp224.h"

#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>

PDataCache pDataCache = NULL;

//Middleware must handle the sharedConnections for TLP224 (when card simulator is used)
DWORD share = SCARD_SHARE_EXCLUSIVE;

/* Entry points for Communication */

/* design decisions:
1. This reference implementation does not provide for any connection types other than generic PC/SC connectivity.
2. If a list of readers is requested, this funciton returns them in a doubly-null-terminated string list. 
*/
DYNAMIC_LIB PIV_RV pivConnect (PIV_Bool sharedConnection, PIV_Byte *connectionDescription, PIV_ULong32 *pCDLength, PIV_CARDHANDLE *pCardHandle) {

	int i;
	PTLV *tlvs;
	int rv; 
	unsigned int tlvcount=0;

	//a wchar_t is 2 bytes on MS, and 4 bytes on most versions
	//of Unix and Linux (because they use 4-byte Unicode).
	//Use LPTSTR not LPSTR for Linux portability 
	LPTSTR readers;
	DWORD readersLen;
	DWORD protocol;

	if (pDataCache == NULL)
		initializeCache(&pDataCache);

	if (sharedConnection)
		share = SCARD_SHARE_SHARED;
	else
		share = SCARD_SHARE_EXCLUSIVE;

	rv = SCardHookEstablish();

	if (rv != PIV_OK) 
	{
		//return rv;
		return PIV_CONNECTION_FAILURE;
	}
	
	if (*pCDLength < 7)
		return PIV_CONNECTION_DESCRIPTION_MALFORMED;

	if (connectionDescription[0]!=0x7F && connectionDescription[1]!=0x21)
		return PIV_CONNECTION_DESCRIPTION_MALFORMED;

	rv = parseTLVTLV(connectionDescription+3,connectionDescription[2],NULL,&tlvcount);

	if (rv!=TLV_SUCCESS) {
		return PIV_CONNECTION_DESCRIPTION_MALFORMED;
	}
	if (tlvcount != 2)
		return PIV_CONNECTION_DESCRIPTION_MALFORMED;
    	
	tlvs = (PTLV*)calloc(sizeof(PTLV),tlvcount);
	parseTLVTLV (connectionDescription+3,connectionDescription[2],tlvs,&tlvcount);

	for (i=0; i<2; i++) 
	{
		switch (tlvs[i]->tag) 
		{
			case 0x81:
				if (tlvs[i]->length == 0) { /* list readers */
					rv = SCardHookListReaders(NULL, &readersLen);
					if (rv != SCARD_E_INSUFFICIENT_BUFFER && rv != SCARD_S_SUCCESS) 
					{
						freeTLVArray(tlvs, 2);
						free(tlvs);
						if(rv == SCARD_E_UNKNOWN_READER || rv  == SCARD_E_DUPLICATE_READER)
							return PIV_CONNECTION_DESCRIPTION_MALFORMED;
						//return rv;
						return PIV_CONNECTION_FAILURE;
					}
					readers = (LPTSTR)calloc(sizeof(char),readersLen);
					rv = SCardHookListReaders(readers, &readersLen);
					if (*pCDLength >= readersLen) 
					{
						memcpy (connectionDescription, readers, readersLen);
						*pCDLength = readersLen;
						free(readers);
						freeTLVArray(tlvs, 2);
						free(tlvs);
						return PIV_OK;
					}
					else {
						*pCDLength = readersLen;
						free(readers);
						freeTLVArray(tlvs, 2);
						free(tlvs);
						return PIV_OK;
					}
				}
				else {
					tlvs[i]->value[tlvs[i]->length] = 0;
					rv = SCardHookConnect((LPCSTR)tlvs[i]->value, share, SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1, (LPSCARDHANDLE)pCardHandle, &protocol);
					if (rv != PIV_OK) 
					{ 
						freeTLVArray(tlvs, 2);
						free(tlvs);
						if (rv == SCARD_E_SHARING_VIOLATION) 
							return PIV_CONNECTION_LOCKED;
                        return PIV_CONNECTION_FAILURE;
					}
				} 
				break;
			case 0x82:
			case 0x83:/* the only connection type supported in this version of the middleware is PC/SC on the local host*/
			case 0x84:
			case 0x85:
			case 0x86:
				if (tlvs[i]->length == 0) {
					*pCDLength=0;
					freeTLVArray(tlvs, 2);
					free(tlvs);
					return PIV_OK;
				}
				freeTLVArray(tlvs, 2);
				free(tlvs);
				return PIV_CONNECTION_FAILURE;
			case 0x90:
				break;
			case 0x91:
			case 0x92:/* Only local PC/SC connections are supported with this version of the middleware */
			case 0x93:/* The card Simulator portion of the Ref. Imp. is treated as PC/SC */
				 continue;
		}
	}
	freeTLVArray(tlvs, 2);
	free(tlvs);
	return PIV_OK;
}

DYNAMIC_LIB PIV_RV pivDisconnect (PIV_CARDHANDLE cardHandle) {
	PIV_RV rv = SCardHookDisconnect((SCARDHANDLE)cardHandle);

	if (rv == SCARD_S_SUCCESS)
		return PIV_OK;
	if (rv == SCARD_W_UNPOWERED_CARD || rv == SCARD_W_RESET_CARD || rv == SCARD_W_REMOVED_CARD)
		return PIV_CARD_READER_ERROR;
#ifdef WIN32
	if ((rv == ERROR_INVALID_HANDLE) || (rv == SCARD_E_INVALID_HANDLE))
		return PIV_INVALID_CARD_HANDLE;
#else
	if (rv == SCARD_E_INVALID_HANDLE)
		return PIV_INVALID_CARD_HANDLE;
#endif
	if (rv == SCARD_E_READER_UNAVAILABLE)
		return PIV_CARD_READER_ERROR;
	if ( rv == SCARD_F_COMM_ERROR)
		return PIV_CARD_READER_ERROR;
	//return PIV_OTHER_CARD_ERROR; //Not such error code in 800-73-1
	return PIV_CARD_READER_ERROR;
}


/*  Entry Points for Data Access */
DYNAMIC_LIB PIV_RV pivSelectCardApplication (PIV_CARDHANDLE cardHandle, PIV_Byte *applicationAID, PIV_ULong32 aidLength, PIV_Byte *applicationProperties, PIV_ULong32 *pAPLength) {
	PIV_RV rv;

	PIV_Byte *buffer;
	PIV_ULong32 bufferLength;
	PIV_Byte aidByteLength;

	if (pDataCache != NULL) {
		if (pDataCache->cardHandle == cardHandle && pDataCache->functionCode == PIV_SELECT_CARD_APPLICATION)
			if (pDataCache->idLen == aidLength)
				if (compare(applicationAID, pDataCache->identifier, aidLength)==0) {
					if (applicationProperties != NULL && *pAPLength >= pDataCache->dataLen) {
						memcpy (applicationProperties, pDataCache->data, pDataCache->dataLen);
						*pAPLength = pDataCache->dataLen;
						clearCache(pDataCache);
						return PIV_OK;
					}
					else {
						*pAPLength = pDataCache->dataLen;
						return PIV_INSUFFICIENT_BUFFER;
					}
				}
	}
	clearCache(pDataCache);

	aidByteLength = aidLength > 255 ? 255 : (PIV_Byte) aidLength;	// limit app ID length to 255 bytes

	rv = StatusWord2RV(pivCardCmdSelectApplication(cardHandle, applicationAID, aidByteLength, NULL, &bufferLength));

	if (rv != PIV_OK)
	{
		//return rv;
		if (rv == SCARD_E_INVALID_HANDLE || rv == PIV_INVALID_CARD_HANDLE)
			return PIV_INVALID_CARD_HANDLE;
		else if (rv == PIV_CARD_APPLICATION_NOT_FOUND)
			return rv; //800-73-1 "an event has occurred..."		
		else 
			return PIV_CARD_READER_ERROR;
	}
	buffer = (PIV_Byte*)calloc(sizeof(PIV_Byte),bufferLength);
	rv = StatusWord2RV(pivCardCmdSelectApplication(cardHandle, applicationAID, aidByteLength, buffer, &bufferLength));

	if (applicationProperties != NULL && *pAPLength >= bufferLength) {
		memcpy (applicationProperties, buffer, bufferLength);
		*pAPLength = bufferLength;
		free(buffer);
		return PIV_OK;
	}
	*pAPLength = bufferLength;

	populateCache(&pDataCache, cardHandle, PIV_SELECT_CARD_APPLICATION, applicationAID, aidLength, buffer, bufferLength);

	free(buffer);
	return PIV_INSUFFICIENT_BUFFER;
}

DYNAMIC_LIB PIV_RV pivLogIntoCardApplication (PIV_CARDHANDLE cardHandle, PIV_Byte *authenticators, PIV_ULong32 authenticatorsLength) {
	PIV_ULong32 offset=0;
	PIV_Byte keyRef;
	PIV_Byte refData[512];
	PIV_Byte refDataLen=0;
	PIV_ULong32 lenlen, authlen=0;
	PIV_ULong32 i;
	PIV_PIN pin;
	PIV_RV rv;

	while (offset < authenticatorsLength) {
		if (authenticators[offset] != 0x67) {
			return PIV_AUTHENTICATOR_MALFORMED;
		}
		if (authenticators[++offset] & 0x80) {
			lenlen= authenticators[offset++] & 0x7f;
			for ( ; lenlen>0; lenlen--)
				authlen = (authlen<<=8) + authenticators[offset++];
		}
		else {
			authlen = authenticators[offset++];
		}
		if (authlen+offset > authenticatorsLength) 
			return PIV_AUTHENTICATOR_MALFORMED;
		keyRef=0xff;
		refDataLen=0;
		while (keyRef==0xff || refDataLen==0) {
			if (authenticators[offset] == 0x83) {
				if (keyRef != 0xff)
					return PIV_AUTHENTICATOR_MALFORMED;
				keyRef = authenticators[offset+2];
				offset+=3;
			}
			else if (authenticators[offset] == 0x81) {
				if (refDataLen != 0)
					return PIV_AUTHENTICATOR_MALFORMED;
				offset++;
				if (authenticators[offset] & 0x80) {
					lenlen= authenticators[offset++] & 0x7f;
					for ( ; lenlen>0; lenlen--)
						refDataLen = (refDataLen<<=8) + authenticators[offset++];
				}
				else {
					refDataLen = authenticators[offset];
					offset++;
				}
				memcpy(refData, authenticators+offset, refDataLen);
				offset+=refDataLen;
			}
			else {
				return PIV_AUTHENTICATOR_MALFORMED;
			}
		}

		if (keyRef == 0 || keyRef == 0x80) {
			if (refDataLen > 8)
				return PIV_AUTHENTICATOR_MALFORMED;
			for (i=0; i<refDataLen; i++)
				pin.PIN[i] = refData[i];
			for (; i<8; i++)
				pin.PIN[i] = 0xff;				
			rv = StatusWord2RV(pivCardCmdVerifyPin(cardHandle, keyRef, &pin));
			if (rv != PIV_OK)
				return rv;
		}
		else  {
			rv = StatusWord2RV(pivCrypt(cardHandle, (PIV_Byte)0x06, keyRef, refData, refDataLen, NULL, &lenlen));
			if (rv != PIV_OK)
			{
				if ( rv == PIV_SW_INCORRECT_PARAMETER_IN_DATA_FIELD)
					return PIV_AUTHENTICATOR_MALFORMED;
				else if ( rv == PIV_AUTHENTICATION_FAILURE )
					return rv;
				else
					return PIV_CARD_READER_ERROR; // for any event that has occurred... 800-73-1
			}
		}
		keyRef = 0xff;
		refDataLen = 0;
	}
	return PIV_OK;
}

DYNAMIC_LIB PIV_RV pivGetData (PIV_CARDHANDLE cardHandle, PIV_Byte *OID, PIV_ULong32 oidLength, PIV_Byte *data, PIV_ULong32 *pDataLength) {
	PIV_Byte tag[3] = {0x5F, 0xC1, 0};
	PIV_ULong32 length=CACHE_SIZE;
	PIV_RV rv = 0;
	const char* oid = (const char*)OID;
	PIV_Byte *buffer;
	PIV_ULong32 bufferLength;

	if (strcmp (oid, "2.16.840.1.101.3.7.1.219.0")==0)
		tag[2] = (PIV_Byte)7;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.48.0")==0)
		tag[2] = (PIV_Byte)2;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.1.1")==0)
		tag[2] = (PIV_Byte)5;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.96.16")==0)
		tag[2] = (PIV_Byte)3;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.96.17")==0)
		tag[2] = (PIV_Byte)4;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.48.1")==0)
		tag[2] = (PIV_Byte)9;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.96.48")==0)
		tag[2] = (PIV_Byte)8;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.1.0")==0)
		tag[2] = (PIV_Byte)0x0A;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.1.2")==0)
		tag[2] = (PIV_Byte)0x0B;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.5.0")==0)
		tag[2] = (PIV_Byte)1;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.144.0")==0)
		tag[2] = (PIV_Byte)6;
	else return PIV_INVALID_OID;

	if (pDataCache != NULL) 
	{
		if (pDataCache->cardHandle == cardHandle && pDataCache->functionCode == PIV_GET_DATA)
			if (pDataCache->idLen == 3)
				if (compare(tag, pDataCache->identifier, 3)==0) {
					//The pDataCache->data contains Tag-Length-Value
					//The conformance test expects data(==value) only!
					if (pDataCache->data[0] == 0x53) {
						if(pDataCache->data[1] == 0x82 ) {
							if( data != NULL && *pDataLength >= pDataCache->dataLen-4){
								memcpy (data, pDataCache->data+4, pDataCache->dataLen-4);
								*pDataLength = pDataCache->dataLen-4;
								clearCache(pDataCache);
								return PIV_OK;
							}
						}
						else {
								if( data != NULL && *pDataLength >= pDataCache->dataLen-2){
								memcpy (data, pDataCache->data+2, pDataCache->dataLen-2);
								*pDataLength = pDataCache->dataLen-2;
								clearCache(pDataCache);
								return PIV_OK;
							}
						}
					} //0x53
					else if (data != NULL && *pDataLength >= pDataCache->dataLen) {
						memcpy (data, pDataCache->data, pDataCache->dataLen);
						*pDataLength = pDataCache->dataLen;
						clearCache(pDataCache);
						return PIV_OK;
					}
					else {
						*pDataLength = pDataCache->dataLen;
						return PIV_INSUFFICIENT_BUFFER;
					}
				}
	}

	clearCache(pDataCache);

	rv = StatusWord2RV(pivCardCmdGetData(cardHandle, tag, 3, NULL, &bufferLength));
	if (rv != PIV_OK) {
		if (rv == PIV_SW_APPLICATION_NOT_FOUND || rv == PIV_CARD_APPLICATION_NOT_FOUND )
			return PIV_DATA_OBJECT_NOT_FOUND; // this is the expected code for above conditions...
		if (rv == PIV_INVALID_CARD_HANDLE || rv == PIV_INVALID_OID || rv == PIV_SECURITY_CONDITIONS_NOT_SATISFIED)
				return rv;
		else
			return PIV_CARD_READER_ERROR; //an event that prevents communication
	}
	buffer = (PIV_Byte*)calloc(sizeof(PIV_Byte),bufferLength);
    rv = StatusWord2RV(pivCardCmdGetData(cardHandle, tag, 3, buffer, &bufferLength));

	//The buffer contains Tag-Length-Value
	//The conformance test expects data(==value) only!
	if (buffer[0] == 0x53) {
		if(buffer[1] == 0x82 ) {
			if( data != NULL && *pDataLength >= bufferLength-4){
				memcpy (data, buffer+4, bufferLength-4);
				*pDataLength = bufferLength-4;
				free(buffer);
				return PIV_OK;
			}
		}
		else {
			if( data != NULL && *pDataLength >= bufferLength-2){
				memcpy (data, buffer+2, bufferLength-2);
				*pDataLength = bufferLength-2;
				free(buffer);
				return PIV_OK;
			}
		}
	} //0x53
	else if (data != NULL && *pDataLength >= bufferLength) {
		memcpy (data, buffer, bufferLength);
		*pDataLength = bufferLength;
		free(buffer);
		return PIV_OK;
	}

	if (buffer[0] == 0x53) {
		if(buffer[1] == 0x82 ) {
			*pDataLength = bufferLength-4;
			populateCache(&pDataCache, cardHandle, PIV_GET_DATA, tag, 3, buffer+4, bufferLength-4);
		}
		else {
			*pDataLength = bufferLength-2;
			populateCache(&pDataCache, cardHandle, PIV_GET_DATA, tag, 3, buffer+2, bufferLength-2);
		}
	} 
	else {
		*pDataLength = bufferLength;
			populateCache(&pDataCache, cardHandle, PIV_GET_DATA, tag, 3, buffer, bufferLength);
	}

	free(buffer);
	return PIV_INSUFFICIENT_BUFFER;

}

DYNAMIC_LIB PIV_RV pivLogoutOfCardApplication (PIV_CARDHANDLE cardHandle) {
	DWORD dwActiveProtocol;
	PIV_RV rv = SCardHookReconnect(cardHandle, share, SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1, SCARD_RESET_CARD, &dwActiveProtocol );

	//SCardHookReconnect returns SCARD_E_xxx errors and PIV_xx? It returns 0 for OK..
	if ( rv == 0x6A82 || rv == 0x9000 || rv == PIV_OK )
		return PIV_OK;

#ifdef WIN32
	if (rv == ERROR_INVALID_HANDLE || rv == SCARD_E_INVALID_HANDLE || rv == PIV_INVALID_CARD_HANDLE)
		return PIV_INVALID_CARD_HANDLE;
#else
	if (rv == SCARD_E_INVALID_HANDLE || rv == PIV_INVALID_CARD_HANDLE)
		return PIV_INVALID_CARD_HANDLE;
#endif
	else
		return PIV_CARD_READER_ERROR; //800-73-1 for any event that prevens communication...
}


/*  Entry Points for Cryptographic Operations */
DYNAMIC_LIB PIV_RV pivCrypt (PIV_CARDHANDLE cardHandle, PIV_Byte algID, PIV_Byte keyReference, PIV_Byte *algInput, PIV_ULong32 inputLength, PIV_Byte *algOutput, PIV_ULong32 *pOutputLength) {
	PIV_ULong32 length=CACHE_SIZE;
	PIV_RV rv;

	PIV_Byte *buffer;
	PIV_ULong32 bufferLength;

	if (pDataCache != NULL) {
		if (pDataCache->cardHandle == cardHandle && pDataCache->functionCode == PIV_CRYPT)
			if (pDataCache->idLen == inputLength)
				if (compare(algInput, pDataCache->identifier, inputLength)==0) {
					if (algOutput != NULL && *pOutputLength >= pDataCache->dataLen) {
						memcpy (algOutput, pDataCache->data, pDataCache->dataLen);
						*pOutputLength = pDataCache->dataLen;
						clearCache(pDataCache);
						return PIV_OK;
					}
					else {
						*pOutputLength = pDataCache->dataLen;
						return PIV_INSUFFICIENT_BUFFER;
					}
				}
	}

	clearCache(pDataCache);

	rv = StatusWord2RV(pivCardCmdGeneralAuthenticate(cardHandle, algID, keyReference, algInput, inputLength, NULL, &bufferLength));
	if (rv != PIV_OK)
	{
		if(rv == PIV_SW_INCORRECT_PARAMETER_IN_P1_OR_P2)
			return PIV_INVALID_KEYREF_OR_ALGORITHM;
		else if (rv == PIV_SW_INCORRECT_PARAMETER_IN_DATA_FIELD)
					return PIV_INPUT_BYTES_MALFORMED;
		else if (rv == PIV_SECURITY_CONDITIONS_NOT_SATISFIED)
			return PIV_SECURITY_CONDITIONS_NOT_SATISFIED;
		else if ( rv == PIV_INVALID_CARD_HANDLE )
			return rv;
		else
			return PIV_CARD_READER_ERROR; // an eventoccurred that prevented communication
	}

	buffer = (PIV_Byte*)calloc(sizeof(PIV_Byte),bufferLength);
    rv = StatusWord2RV(pivCardCmdGeneralAuthenticate(cardHandle, algID, keyReference, algInput, inputLength , buffer, &bufferLength));

	if (algOutput != NULL && *pOutputLength >= bufferLength) {
		memcpy (algOutput, buffer, bufferLength);
		*pOutputLength = bufferLength;
		free(buffer);
		return PIV_OK;
	}
	*pOutputLength = bufferLength;

	populateCache(&pDataCache, cardHandle, PIV_CRYPT, &keyReference, 1, buffer, bufferLength);
	
	free(buffer);
	return PIV_INSUFFICIENT_BUFFER;
}

/* Entry Points for Credential Initialization and Administration */
DYNAMIC_LIB PIV_RV pivPutData (PIV_CARDHANDLE cardHandle, PIV_Byte *OID, PIV_ULong32 oidLength, PIV_Byte *data, PIV_ULong32 dataLength) {
	PIV_Byte tag[3] = {0x5F, 0xC1, 0};
	PIV_ULong32 length=CACHE_SIZE;
	PIV_RV rv;
	const char* oid = (const char*)OID;
	
	PIV_ULong32 bufferLength;

	if (strcmp (oid, "2.16.840.1.101.3.7.1.219.0")==0)
		tag[2] = (PIV_Byte)7;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.48.0")==0)
		tag[2] = (PIV_Byte)2;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.1.1")==0)
		tag[2] = (PIV_Byte)5;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.96.16")==0)
		tag[2] = (PIV_Byte)3;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.96.17")==0)
		tag[2] = (PIV_Byte)4;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.48.1")==0)
		tag[2] = (PIV_Byte)9;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.96.48")==0)
		tag[2] = (PIV_Byte)8;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.1.0")==0)
		tag[2] = (PIV_Byte)0x0A;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.1.2")==0)
		tag[2] = (PIV_Byte)0x0B;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.5.0")==0)
		tag[2] = (PIV_Byte)1;
	else if (strcmp (oid, "2.16.840.1.101.3.7.2.144.0")==0)
		tag[2] = (PIV_Byte)6;
	else return PIV_INVALID_OID;

	rv = StatusWord2RV(pivCardCmdPutData(cardHandle, tag, 3, data, dataLength,NULL, &bufferLength));

	if (rv == SCARD_E_INVALID_HANDLE || rv == PIV_INVALID_CARD_HANDLE)
		return PIV_INVALID_CARD_HANDLE;
	else if ( rv == PIV_OK || rv == PIV_SECURITY_CONDITIONS_NOT_SATISFIED)
		return rv;
	else if ( rv == 0x6A84 )
		return PIV_INSUFFICIENT_CARD_RESOURCE;
	else
		return PIV_CARD_READER_ERROR;
}

DYNAMIC_LIB PIV_RV pivGenerateKeyPair (PIV_CARDHANDLE cardHandle, PIV_Byte keyReference, PIV_Byte cryptographicMechanism, PIV_Byte *publicKey, PIV_ULong32 *pKeyLength) {
	PIV_RV rv;
	StatusWord sw;

	PIV_Byte *buffer;
	PIV_ULong32 bufferLength;

	if (pDataCache != NULL) {
		if (pDataCache->cardHandle == cardHandle && pDataCache->functionCode == PIV_GENERATE_KEY_PAIR)
			if (pDataCache->idLen == 1)
				if (compare(&keyReference, pDataCache->identifier, 1)==0) {
					if (publicKey != NULL && *pKeyLength >= pDataCache->dataLen) {
						memcpy (publicKey, pDataCache->data, pDataCache->dataLen);
						*pKeyLength = pDataCache->dataLen;
						clearCache(pDataCache);
						return PIV_OK;
					}
					else {
						*pKeyLength = pDataCache->dataLen;
						return PIV_INSUFFICIENT_BUFFER;
					}
				}
	}

	clearCache(pDataCache);

	sw = pivCardCmdGenerateAsymmetricKeyPair(cardHandle, keyReference, cryptographicMechanism, NULL, 0, NULL, &bufferLength);
	if (sw == PIV_SW_INCORRECT_PARAMETER_IN_P1_OR_P2)
		return PIV_INVALID_KEY_OR_KEYALG_COMBINATION;
	if (sw == PIV_SW_INCORRECT_PARAMETER_IN_DATA_FIELD)
		return PIV_UNSUPPORTED_CRYPTOGRAPHIC_MECHANISM;

	rv = StatusWord2RV(sw);
	if (rv != PIV_OK) {
		if( rv == PIV_SECURITY_CONDITIONS_NOT_SATISFIED )
			return rv;
		else if ( rv == SCARD_E_INVALID_HANDLE || rv == PIV_INVALID_CARD_HANDLE )
			return PIV_INVALID_CARD_HANDLE;
		else
			return PIV_CARD_READER_ERROR; // an event that prevents communication...
	}
	buffer = (PIV_Byte*)calloc(sizeof(PIV_Byte),bufferLength);
    rv = StatusWord2RV(pivCardCmdGenerateAsymmetricKeyPair(cardHandle, keyReference, cryptographicMechanism, NULL, 0, buffer, &bufferLength));

	if (publicKey != NULL && *pKeyLength >= bufferLength) {
		memcpy (publicKey, buffer, bufferLength);
		*pKeyLength = bufferLength;
		free(buffer);
		return PIV_OK;
	}
	*pKeyLength = bufferLength;

	populateCache(&pDataCache, cardHandle, PIV_GENERATE_KEY_PAIR, &keyReference, 1, buffer, bufferLength);
	
	free(buffer);
	return PIV_INSUFFICIENT_BUFFER;
}

DYNAMIC_LIB PIV_RV StatusWord2RV (StatusWord sw) {
	if (((sw&0xff00)==0x6c00) || ((sw&0xfff0)==PIV_SW_VERIFY_FAILED))
		return PIV_AUTHENTICATION_FAILURE;

	switch (sw) {
		case PIV_SW_SUCCESS: 
			return PIV_OK;
		case PIV_SW_SECURITY_CONDITION_NOT_SATISFIED:
		case PIV_SW_CONDITIONS_OF_USE_NOT_SATISFIED:
			return PIV_SECURITY_CONDITIONS_NOT_SATISFIED;
		
//		case PIV_SW_INCORRECT_PARAMETER_IN_DATA_FIELD: //can not map it here because pivCrypt and pivLogIntoCard.. maps it differently
		case PIV_SW_KEY_REFERENCE_NOT_FOUND:
			return PIV_AUTHENTICATOR_MALFORMED; //PIV_INPUT_BYTES_MALFORMED;//
		
		case PIV_SW_APPLICATION_NOT_FOUND:
			return PIV_CARD_APPLICATION_NOT_FOUND;
		
//		case PIV_SW_KEY_REFERENCE_NOT_FOUND:
//			return PIV_AUTHENTICATION_FAILURE; // in 800-73-1 6A88 is mapped to PIV_AUTHENTICATOR_MALFORMED

		case PIV_SW_NOT_ENOUGH_MEMORY:
			return PIV_INSUFFICIENT_CARD_RESOURCE;

		default: return sw;
	}

}

DYNAMIC_LIB void dumpData(char* memAddr,unsigned long len)
{
	unsigned long i = 0;
	printf("Number of bytes %d\n", len);

	for (i=0; i < len; i++) {
		printf("%02X", memAddr[i] & 0xff);

		if (i != 0) {
			if((i +1) % 4 == 0)
				printf(" ");
			if((i +1) % 16 == 0) 
				printf("\n");
		}
	}
}
