// pivCardEdge: ANSI C implementation of the API in SP 800-73, Chapter 7 
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The PIV Middleware Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include "pivCardEdge.h"
#include "pivdefs.h"
#include "pivpcsc.h"
#include "pivtlp224.h"

#ifdef _DEBUG
#include <stdio.h>
#endif

extern PDataCache pDataCache;

void clearCache(PDataCache pCache) {
	if(pCache != NULL) {
		pCache->idLen=0;
		pCache->dataLen=0;
		pCache->cardHandle=0;
		pCache->functionCode=0;
	}
}

void initializeCache(PDataCache *ppDataCache) {
	if (*ppDataCache == NULL)
		*ppDataCache = (DataCache*)calloc(sizeof(DataCache),1);
	clearCache(*ppDataCache);
}
void freeCache(PDataCache *ppDataCache) {
	if (*ppDataCache != NULL) {
		free(*ppDataCache);
		*ppDataCache=NULL;
	}
}

void populateCache (PDataCache *ppDataCache, PIV_CARDHANDLE cardHandle, PIV_Byte functionCode, PIV_Byte *identifier, PIV_ULong32 idLen, PIV_Byte* data, PIV_ULong32 dataLength) {
	initializeCache(ppDataCache);
	(*ppDataCache)->cardHandle = cardHandle;
	(*ppDataCache)->functionCode = functionCode;
	(*ppDataCache)->idLen = idLen;
	(*ppDataCache)->dataLen = dataLength;
	memcpy((*ppDataCache)->data, data, dataLength);
	memcpy((*ppDataCache)->identifier, identifier, idLen);
}

int compare (PIV_Byte* b1, PIV_Byte* b2, PIV_ULong32 len) {
	unsigned int i;
	for (i=0; i<len; i++)
		if (b1[i]!=b2[i])
			return 1;
	return 0;
}

PIV_RV PCSC_To_PIV (PIV_ULong32 rv) {
	switch (rv) {
		case SCARD_S_SUCCESS: 
			return PIV_OK;
#ifdef WIN32
		case ERROR_INVALID_HANDLE:
#endif
		case SCARD_E_INVALID_HANDLE: 
			return PIV_INVALID_CARD_HANDLE;
		
		default: return rv;
	}
}

StatusWord ReadStatusWord(PCSCLPTransmitBuffer txbuff) {
	return (((StatusWord)txbuff->apduResponse[txbuff->apduResponseSize-2])<<8) +txbuff->apduResponse[txbuff->apduResponseSize-1];
}

StatusWord GetResponse (PIV_CARDHANDLE cardHandle, PIV_Byte le, PIV_Byte *buffer, PIV_ULong32 *pBufferLength) {

	PIV_RV rv;
	PCSCTransmitBuffer transmitBuffer;
	unsigned long sw=0;
	transmitBuffer.apduResponseSize = PCSC_MAX_BUFFERSIZE;

	if (buffer==NULL || *pBufferLength < le) {
		*pBufferLength = le;
		return PIV_INSUFFICIENT_BUFFER;
	}
	
	/* Get Response APDU */
	transmitBuffer.buffer[APDU_CLA] = 0x00;
	transmitBuffer.buffer[APDU_INS] = 0xC0;
	transmitBuffer.buffer[APDU_P1]  = 0x00;
	transmitBuffer.buffer[APDU_P2]  = 0x00;
	transmitBuffer.buffer[APDU_P3]  = le;
	transmitBuffer.bufferSize = 5;
	
	rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);

	if (rv != SCARD_S_SUCCESS) 
	{
#ifdef _DEBUG
		printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error( rv));
#endif
		return PCSC_To_PIV(rv);
	}
	sw = ReadStatusWord(&transmitBuffer);
	
	if ((sw&0xFF00) == 0x6C00) {
		transmitBuffer.buffer[APDU_P3]=(PIV_UChar8)sw;
		// trying to fix stuff...
		transmitBuffer.apduResponseSize=PCSC_MAX_BUFFERSIZE;
		rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
		if (rv != SCARD_S_SUCCESS) {
#ifdef _DEBUG
			printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
			return PCSC_To_PIV(rv);
		}
		sw = ReadStatusWord(&transmitBuffer);
	}

	if (sw != 0x9000 && (sw & 0xFF00) != 0x6100) return sw;
    memcpy (buffer, &(transmitBuffer.apduResponse[0]), transmitBuffer.apduResponseSize-2);
	return sw;
}

unsigned int GetAllResponses (PIV_CARDHANDLE cardHandle, unsigned int status) {
	unsigned long len;
	PIV_UShort16 bytesToRead;
	PIV_Byte le;

	while ( (status&0xff00) == 0x6100) {
		le = (PIV_Byte)status&0x0ff;
		//if (le == 0xff)
		//	le = 0;
		bytesToRead = (le == 0) ? 256 : le;
		len = CACHE_SIZE - pDataCache->dataLen;
		pDataCache->dataLen += bytesToRead;
		status = GetResponse(cardHandle, le, &(pDataCache->data[0])+pDataCache->dataLen-bytesToRead, &len);
	}
	return pDataCache->dataLen;
}

int berGenerateLength (PIV_ULong32 len, PIV_Byte *buffer, PIV_ULong32 *pBufferLength) {
	PIV_ULong32 l;
	if (len < 0x80) {
		buffer[0] = (PIV_Byte)len;
		*pBufferLength = 1;
		return 0;
	}
    l = len;
	*pBufferLength = 0;
	while (l>0) {
		l >>= 8;
		(*pBufferLength)++;
	}

	buffer[0] = 0x80 | (0xff & ((PIV_Byte)*pBufferLength));
    
	for (l=*pBufferLength; l>0; l--) {
		buffer[l] = (PIV_Byte)(len&0xff);
		len >>= 8;
	}

	(*pBufferLength)++;
	return 0;
}

StatusWord 
pivCardCmdSelectApplication (PIV_CARDHANDLE cardHandle, PIV_Byte *aid, PIV_Byte aidLength, PIV_Byte *applicationPropertyTemplate, PIV_ULong32 *pPropertyTemplateLength)
{
	PIV_RV rv;
	PCSCTransmitBuffer transmitBuffer;
	PIV_ULong32 status;

	transmitBuffer.apduResponseSize = PCSC_MAX_BUFFERSIZE;

	/* if the cache is full, check to see if the request is the same.   
	 * if so, copy the data into the buffer and return. */

	/* Warning!!
	 * There's a slight problem with the following block of code which is manifest in each card 
	 * edge function that might have return data.  Errors may occur if multiple applications are 
	 * communicating with a card at the same time, or if functions from alternative smart card/pcsc 
	 * libraries are called.  The problem is associated with the return-data caching mechanism in 
	 * place here.  When a command returns data and there is insufficient space in the 
	 * return-buffer parameter, the data is put into a cache to be returned by a subsequent call 
	 * of the same command with the same parameters excepting that a sufficiently-large buffer is 
	 * provided.  If a command in this library is called that creates a state-change, the buffer 
	 * will be emptied and things will continue fine.  However, if a different library call 
	 * changes the state of the card, this library will continue as if there were no change, 
	 * which might create unexpected and/or incorrect results.  
     */
	if (pDataCache != NULL) {
		if (pDataCache->cardHandle == cardHandle && pDataCache->functionCode == SELECT_APPLICATION)
			if (pDataCache->idLen == aidLength)
				if (compare(aid, pDataCache->identifier, aidLength)==0) {
					if (applicationPropertyTemplate != NULL && *pPropertyTemplateLength >= pDataCache->dataLen) {
						memcpy (applicationPropertyTemplate, pDataCache->data, pDataCache->dataLen);
						*pPropertyTemplateLength = pDataCache->dataLen;
						clearCache(pDataCache);
						return PIV_SW_SUCCESS;
					}
					else {
						*pPropertyTemplateLength = pDataCache->dataLen;
						return PIV_SW_SUCCESS;
					}
				}
	}
	/* the cache has been used, or we've changed commands, so go ahead and purge it. */
	clearCache(pDataCache);
	
	/* Select Applet APDU */
	transmitBuffer.buffer[APDU_CLA] = 0x00;
	transmitBuffer.buffer[APDU_INS] = 0xA4;
	transmitBuffer.buffer[APDU_P1]  = 0x04;
	transmitBuffer.buffer[APDU_P2]  = 0x00;
	transmitBuffer.buffer[APDU_P3]  = aidLength;

	memcpy (transmitBuffer.buffer+APDU_DATA, aid, aidLength);
	transmitBuffer.bufferSize = 5+aidLength;
	rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);

	if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
		printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
		return PCSC_To_PIV(rv);
	}

	status = ReadStatusWord(&transmitBuffer);
    
	if (status!=PIV_SW_SUCCESS && (status&0xff00)!=PIV_SW_SUCCESS_DATA)
		return status;

	populateCache(&pDataCache, cardHandle, SELECT_APPLICATION, aid, aidLength, transmitBuffer.apduResponse, transmitBuffer.apduResponseSize-2);

	GetAllResponses (cardHandle, status);

	/* see if there's room to return the data, else cache it. 
	* 
	*  if the return buffer is not NULL and the provided length 
	*  is large enough, fill buffer and return success */
	if (applicationPropertyTemplate != NULL) {
		if (pPropertyTemplateLength != NULL) {
			if (*pPropertyTemplateLength >= pDataCache->dataLen) {
				memcpy (applicationPropertyTemplate, pDataCache->data, pDataCache->dataLen);
				*pPropertyTemplateLength = pDataCache->dataLen;
				return PIV_SW_SUCCESS;
			}
		}
	}
	/* if you get here, the buffer is NULL or too small, indicate length, and return */
	
	*pPropertyTemplateLength = pDataCache->dataLen;
	return PIV_SW_SUCCESS;
}


StatusWord 
pivCardCmdGetData(PIV_CARDHANDLE cardHandle, PIV_Byte *OID, PIV_Byte oidLength, PIV_Byte *data, PIV_ULong32 *pDataLength) 
{
	PIV_RV rv;
	PCSCTransmitBuffer transmitBuffer;
	PIV_ULong32 status;

	if (pDataCache != NULL) {
		if (pDataCache->cardHandle == cardHandle && pDataCache->functionCode == GET_DATA)
			if (pDataCache->idLen == oidLength)
				if (compare(OID, pDataCache->identifier, oidLength)==0) {
					if (data != NULL && *pDataLength >= pDataCache->dataLen) {
						memcpy (data, pDataCache->data, pDataCache->dataLen);
						*pDataLength = pDataCache->dataLen;
						clearCache(pDataCache);
						return PIV_SW_SUCCESS;
					}
					else {
						*pDataLength = pDataCache->dataLen;
						return PIV_SW_SUCCESS;
					}
				}
	}
	/* the cache has been used, or we've changed commands, so go ahead and purge it. */
	clearCache(pDataCache);

	transmitBuffer.buffer[APDU_CLA] =  0x00;
	transmitBuffer.buffer[APDU_INS] =  0xCB;
	transmitBuffer.buffer[APDU_P1] =   0x3F;
	transmitBuffer.buffer[APDU_P2] =   0xFF;
	transmitBuffer.buffer[APDU_P3] =   oidLength + 2;
	transmitBuffer.buffer[APDU_DATA] = 0x5C;
	transmitBuffer.buffer[APDU_DATA + 1] = oidLength;
	memcpy (transmitBuffer.buffer+APDU_DATA+2, OID, oidLength);

	transmitBuffer.bufferSize = 7+oidLength;
	transmitBuffer.apduResponseSize = PCSC_MAX_BUFFERSIZE;
	rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);

	if (rv != SCARD_S_SUCCESS) {
#ifdef _DEBUG
		printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
		return PCSC_To_PIV(rv);
	}

	status = ReadStatusWord(&transmitBuffer);
	if (status!=PIV_SW_SUCCESS && (status&0xff00)!=PIV_SW_SUCCESS_DATA)
		return status;

	populateCache(&pDataCache, cardHandle, GET_DATA, OID, oidLength, transmitBuffer.apduResponse, transmitBuffer.apduResponseSize-2);
	/* if there's more data, GetResponse until there ain't */
	GetAllResponses (cardHandle, status);

	/* if the return buffer is not NULL and the provided length is large enough, fill buffer and return success */
	if (data != NULL) {
		if (pDataLength != NULL) {
			if (*pDataLength >= pDataCache->dataLen) {
				memcpy (data, pDataCache->data, pDataCache->dataLen);
				*pDataLength = pDataCache->dataLen;
				clearCache(pDataCache);
				return PIV_SW_SUCCESS;
			}
		}
	}

	*pDataLength = pDataCache->dataLen;
	return PIV_SW_SUCCESS;
}

StatusWord 
pivCardCmdVerifyPin(PIV_CARDHANDLE cardHandle, PIV_Byte keyReference, PIV_PIN *pin) 
{
	PIV_RV rv;
	PCSCTransmitBuffer transmitBuffer;

	transmitBuffer.apduResponseSize = PCSC_MAX_BUFFERSIZE;
	freeCache(&pDataCache);

	/* Verify PIN APDU */
	transmitBuffer.buffer[APDU_CLA] = 0x00; 
	transmitBuffer.buffer[APDU_INS] = 0x20; 
	transmitBuffer.buffer[APDU_P1]  = 0x00; 
	transmitBuffer.buffer[APDU_P2]  = 0x80;//keyReference;
	transmitBuffer.buffer[APDU_P3]  = 0x08;
	transmitBuffer.bufferSize = 13;

	memcpy (transmitBuffer.buffer + APDU_DATA, pin->PIN, 8);

	rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
	if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
		printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
		return PCSC_To_PIV(rv);
	}

	return ReadStatusWord(&transmitBuffer);
}

StatusWord
pivCardCmdChangeReferenceData(PIV_CARDHANDLE cardHandle, PIV_Byte keyReference, PIV_RefData *pRefData) 
{
	PIV_RV rv;
	PCSCTransmitBuffer transmitBuffer;
	transmitBuffer.apduResponseSize = PCSC_MAX_BUFFERSIZE;
	freeCache(&pDataCache);

	/* Change Reference Data APDU */
	transmitBuffer.buffer[APDU_CLA] = 0x00; 
	transmitBuffer.buffer[APDU_INS] = 0x24; 
	transmitBuffer.buffer[APDU_P1] =  0x04; 
	transmitBuffer.buffer[APDU_P2] =  keyReference;
	transmitBuffer.buffer[APDU_P3] =  0x10;
	transmitBuffer.bufferSize = 21;

	memcpy (transmitBuffer.buffer+APDU_DATA, pRefData->oldData, 0x10);
	rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);

	if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
		printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
		return PCSC_To_PIV(rv);
	}
	return ReadStatusWord(&transmitBuffer);
}

StatusWord
pivCardCmdResetRetryCounter(PIV_CARDHANDLE cardHandle, PIV_Byte keyReference, PIV_Byte *data) 
{
	PIV_RV rv;
	PCSCTransmitBuffer transmitBuffer;
	transmitBuffer.apduResponseSize = PCSC_MAX_BUFFERSIZE;
	freeCache(&pDataCache);

	/* Reset Retry Counter APDU */
	transmitBuffer.buffer[APDU_CLA] = 0x00; 
	transmitBuffer.buffer[APDU_INS] = 0x2C; 
	transmitBuffer.buffer[APDU_P1] =  0x00; 
	transmitBuffer.buffer[APDU_P2] =  keyReference;
	transmitBuffer.buffer[APDU_P3] =  0x10;

	memcpy (transmitBuffer.buffer+APDU_DATA, data, 0x10);
	rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
	
	if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
		printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
		return PCSC_To_PIV(rv);
	}

	return ReadStatusWord(&transmitBuffer);
}

StatusWord
pivCardCmdGeneralAuthenticate(PIV_CARDHANDLE cardHandle, PIV_Byte algorithmReference, PIV_Byte keyReference, PIV_Byte *cryptoInput, PIV_ULong32 inputLength, PIV_Byte *authData, PIV_ULong32 *pAuthDataLength)
{
	PIV_RV rv;
	PCSCTransmitBuffer transmitBuffer;
	unsigned int offset;
	StatusWord status;

	transmitBuffer.apduResponseSize = PCSC_MAX_BUFFERSIZE;

	if (pDataCache != NULL) {
		if (pDataCache->cardHandle == cardHandle && pDataCache->functionCode == GENERAL_AUTHENTICATE)
			if (pDataCache->idLen == inputLength)
				if (compare(cryptoInput, pDataCache->identifier, inputLength)==0) {
					if (authData != NULL && *pAuthDataLength >= pDataCache->dataLen) {
						memcpy (authData, pDataCache->data, pDataCache->dataLen);
						*pAuthDataLength = pDataCache->dataLen;
						clearCache(pDataCache);
						return PIV_SW_SUCCESS;
					}
					else {
						*pAuthDataLength = pDataCache->dataLen;
						return PIV_SW_SUCCESS;
					}
				}
	}

	/* the cache has been used, or we've changed commands, so go ahead and purge it. */
	clearCache(pDataCache);


	/* General Authenticate APDU */
	transmitBuffer.buffer[APDU_CLA] = 0x00; 
	transmitBuffer.buffer[APDU_INS] = 0x87;
	transmitBuffer.buffer[APDU_P1]  = algorithmReference; 
	transmitBuffer.buffer[APDU_P2]  = keyReference;
		
	if (inputLength < 255) {
		/* no need to chain */
		transmitBuffer.buffer[APDU_P3] = (PIV_Byte)inputLength;
		memcpy (transmitBuffer.buffer+APDU_DATA, cryptoInput, inputLength);
		transmitBuffer.bufferSize = 5+inputLength;
		rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
		if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
			printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif	
			return PCSC_To_PIV(rv);
		}
	}
	else {
		/* need to chain */
		offset=0;
		transmitBuffer.buffer[APDU_CLA] = 0x10;
		transmitBuffer.buffer[APDU_P3] = 200;
		transmitBuffer.bufferSize = 205;
		while (offset + 200 < inputLength) {
			memcpy (transmitBuffer.buffer+APDU_DATA, cryptoInput+offset, 200);
			rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
			if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
				printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
				return PCSC_To_PIV(rv);
			}
			offset += 200;
		}
		transmitBuffer.buffer[APDU_CLA] = 0;
		transmitBuffer.buffer[APDU_P3] = (PIV_Byte)(inputLength-offset);
		transmitBuffer.bufferSize = 5+inputLength-offset;
		memcpy (transmitBuffer.buffer + APDU_DATA, cryptoInput + offset, inputLength-offset);
		rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
		if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
			printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
			return PCSC_To_PIV(rv);
		}
	}

	if (rv != SCARD_S_SUCCESS) {
#ifdef _DEBUG
		printf ("SCardTransmit Failed, returning error code %d", rv);
#endif
		return PCSC_To_PIV(rv);
	}

	status = ReadStatusWord(&transmitBuffer);

	if (status!=0x9000 && (status&0xff00)!=0x6100)
		return status;

	
	populateCache(&pDataCache, cardHandle, GENERAL_AUTHENTICATE, cryptoInput, inputLength, transmitBuffer.apduResponse, transmitBuffer.apduResponseSize-2);

	/* if there's more data, GetResponse it until there ain't */
	GetAllResponses(cardHandle, status);

	/* if the return buffer is not NULL and the provided length is large enough, fill buffer and return success */
	if (authData != NULL) {
		if (pAuthDataLength != NULL) {
			if (*pAuthDataLength >= pDataCache->dataLen) {
				memcpy (authData, pDataCache->data, pDataCache->dataLen);
				*pAuthDataLength = pDataCache->dataLen;
				return PIV_SW_SUCCESS;
			}
		}
	}
	/* if you get here, the buffer is NULL or too small, indicate length, and return */
	
	*pAuthDataLength = pDataCache->dataLen;
	return PIV_SW_SUCCESS;
}

StatusWord
pivCardCmdPutData(PIV_CARDHANDLE cardHandle, PIV_Byte *oid, PIV_ULong32 oidLen, PIV_Byte *data, PIV_ULong32 dataLength, PIV_Byte *authData, PIV_ULong32 *pAuthDataLength) 
{
	PIV_RV rv;
	PCSCTransmitBuffer transmitBuffer;
	StatusWord status;
	PIV_ULong32 offset=0;
	unsigned long len;

	if (pDataCache != NULL) {
		if (pDataCache->cardHandle == cardHandle && pDataCache->functionCode == PUT_DATA)
			if (pDataCache->idLen == dataLength)
				if (compare(data, pDataCache->identifier, dataLength)==0) {
					if (authData != NULL && *pAuthDataLength >= pDataCache->dataLen) {
						memcpy (authData, pDataCache->data, pDataCache->dataLen);
						*pAuthDataLength = pDataCache->dataLen;
						clearCache(pDataCache);
						return PIV_OK;
					}
					else {
						*pAuthDataLength = pDataCache->dataLen;
						return PIV_OK;
					}
				}
	}
	clearCache(pDataCache); 

	transmitBuffer.apduResponseSize = PCSC_MAX_BUFFERSIZE;

	/* Put Data APDU */
	transmitBuffer.buffer[APDU_CLA] = 0x00; 
	transmitBuffer.buffer[APDU_INS] = 0xDB; 
	transmitBuffer.buffer[APDU_P1]  = 0x3F; 
	transmitBuffer.buffer[APDU_P2]  = 0xFF;

	transmitBuffer.buffer[APDU_DATA]   = 0x5C;
	transmitBuffer.buffer[APDU_DATA+1] = 0x03;
	transmitBuffer.buffer[APDU_DATA+2] = oid[0];
	transmitBuffer.buffer[APDU_DATA+3] = oid[1];
	transmitBuffer.buffer[APDU_DATA+4] = oid[2];
	
	transmitBuffer.buffer[APDU_DATA+5] = 0x53;
	berGenerateLength ( dataLength, transmitBuffer.buffer+APDU_DATA+6, &len); 

	if (dataLength < 240-oidLen)
	{
		/* no need to chain */
		//len gives the number of bytes dataLength needs/has or 1 if dataLength < 0x80,
		//and dataLength is a PIV_ULong32 - casting all should not change the value then.
		transmitBuffer.buffer[APDU_P3] = (PIV_Byte)(dataLength + len + 6);
		memcpy (transmitBuffer.buffer+APDU_DATA+len+6, data, dataLength);
		transmitBuffer.bufferSize = 5+dataLength+len+6;
		rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
		if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
			printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
			return PCSC_To_PIV(rv);
		}
		status = ReadStatusWord(&transmitBuffer);
	}
	else {
		/* need to chain */
		offset=0;
		transmitBuffer.buffer[APDU_CLA] = 0x10;
		transmitBuffer.buffer[APDU_P3] = (PIV_Byte)(200+len+6);
		transmitBuffer.bufferSize = 205+len+6;
		memcpy (transmitBuffer.buffer+APDU_DATA+6+len, data+offset, 200);
		rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
		if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
			printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
			return PCSC_To_PIV(rv);
		}
		offset += 200;

		while (offset + 200 < dataLength) {
			transmitBuffer.buffer[APDU_P3] = 200;
			transmitBuffer.bufferSize = 205;
			memcpy (transmitBuffer.buffer+APDU_DATA, data+offset, 200);
			rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
			if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
				printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
				return PCSC_To_PIV(rv);
			}
			offset += 200;
		}
		transmitBuffer.buffer[APDU_CLA] = 0;
		transmitBuffer.buffer[APDU_P3] = (PIV_Byte)(dataLength-offset);
		transmitBuffer.bufferSize = 5+dataLength-offset;
		memcpy (transmitBuffer.buffer + APDU_DATA, data + offset, dataLength-offset);
		rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
		if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
			printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
			return PCSC_To_PIV(rv);
		}
	}

	if (rv != SCARD_S_SUCCESS) {
#ifdef _DEBUG
		printf ("SCardTransmit Failed, returning error code %d", rv);
#endif
		return PCSC_To_PIV(rv);
	}

	status = ReadStatusWord(&transmitBuffer);

	status = ReadStatusWord(&transmitBuffer);
	if (status!=PIV_SW_SUCCESS && (status&0xff00)!=PIV_SW_SUCCESS_DATA)
		return status;

	populateCache(&pDataCache, cardHandle, GENERAL_AUTHENTICATE, data, dataLength, transmitBuffer.apduResponse, transmitBuffer.apduResponseSize-2);

	/* if there's more data, GetResponse it until there ain't */
	GetAllResponses(cardHandle, status);

	/* if the return buffer is not NULL and the provided length is large enough, fill buffer and return success */
	if (authData != NULL) {
		if (pAuthDataLength != NULL) {
			if (*pAuthDataLength >= pDataCache->dataLen) {
				memcpy (authData, pDataCache->data, pDataCache->dataLen);
				*pAuthDataLength = pDataCache->dataLen;
				return PCSC_To_PIV(rv);
			}
		}
	}
	/* if you get here, the buffer is NULL or too small, so cache it, indicate length, and return */
	
	*pAuthDataLength = pDataCache->dataLen;

	return PIV_SW_SUCCESS;
}

StatusWord
pivCardCmdGenerateAsymmetricKeyPair(PIV_CARDHANDLE cardHandle, PIV_Byte keyReference, PIV_Byte cryptoMechanism, PIV_Byte *parameter, PIV_ULong32 parameterLength, PIV_Byte *publicKeyDataObjects, PIV_ULong32 *pPublicKeyDataObjectsLength) 
{
	PIV_RV rv;
	PCSCTransmitBuffer transmitBuffer;
	PIV_ULong32 offset=0;
	unsigned long len=0;
	StatusWord status;
	PIV_Byte *buffer;
	PIV_ULong32 bufferLength;
	PIV_Byte *p;
	PIV_ULong32 pLength;
	

	transmitBuffer.apduResponseSize = PCSC_MAX_BUFFERSIZE;

	if (pDataCache != NULL) {
		if (pDataCache->cardHandle == cardHandle && pDataCache->functionCode == GENERATE_ASYMMETRIC_KEY_PAIR)
			if (pDataCache->idLen == 1)
				if (compare(&keyReference, pDataCache->identifier, 1)==0) {
					if (*pPublicKeyDataObjectsLength >= pDataCache->dataLen) {
						memcpy (publicKeyDataObjects, pDataCache->data, pDataCache->dataLen);
						*pPublicKeyDataObjectsLength = pDataCache->dataLen;
						clearCache(pDataCache);
						return PIV_OK;
					}
					else {
						*pPublicKeyDataObjectsLength = pDataCache->dataLen;
						return PIV_OK;
					}
				}
	}
	clearCache(pDataCache);  

	if (parameterLength == 0) {
		bufferLength = 5;
		buffer = (PIV_Byte*)calloc(sizeof(PIV_Byte),5);
		buffer[0] = 0xAC;
		buffer[1] = 0x03;
		buffer[2] = 0x80;
		buffer[3] = 0x01;
		buffer[4] = cryptoMechanism;
	}
	else { 
		p = (PIV_Byte*)calloc (sizeof(PIV_Byte),(parameterLength + 10));
		p[0] = 0x81;
		berGenerateLength(parameterLength, p+1, &len);
		memcpy(p+1+len, parameter, parameterLength);
		pLength = 1+len+parameterLength;
		buffer = (PIV_Byte*)calloc(sizeof(PIV_Byte),(pLength + 15));
		buffer[0] = 0xAC;
		berGenerateLength(pLength+3, buffer+1, &len);
		bufferLength = 1 + len + 3 + pLength;
		buffer[len+1] = 0x80;
		buffer[len+2] = 0x01;
		buffer[len+3] = cryptoMechanism;
		memcpy(buffer+len+4, p, pLength);
		free(p);
	}

	/* Generate Asymmetric Key Pair APDU */
	transmitBuffer.buffer[APDU_CLA] = 0x00; 
	transmitBuffer.buffer[APDU_INS] = 0x47;
	transmitBuffer.buffer[APDU_P1]  = 0x00;
	transmitBuffer.buffer[APDU_P2]  = keyReference;
	if (bufferLength<255) {
		transmitBuffer.bufferSize = (PIV_Byte)bufferLength + 5;
		transmitBuffer.buffer[APDU_P3] = (PIV_Byte)bufferLength;
		memcpy (transmitBuffer.buffer+APDU_DATA, buffer, bufferLength);
		rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
		if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
			printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif	
			free(buffer);
			return PCSC_To_PIV(rv);
		}
	}

	else {
		/* need to chain */
		offset=0;
		transmitBuffer.buffer[APDU_CLA] = 0x10;
		transmitBuffer.buffer[APDU_P3] = 200;
		transmitBuffer.bufferSize = 205;
		while (offset + 200 < bufferLength) {
			memcpy (transmitBuffer.buffer+APDU_DATA, buffer+offset, 200);
			rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
			if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
				printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
				free(buffer);
				return PCSC_To_PIV(rv);
			}
			offset += 200;
		}
		transmitBuffer.buffer[APDU_CLA] = 0;
		transmitBuffer.buffer[APDU_P3] = (PIV_Byte)(bufferLength-offset);
		transmitBuffer.bufferSize = 5+bufferLength-offset;
		memcpy (transmitBuffer.buffer + APDU_DATA, buffer + offset, bufferLength-offset);
		rv = SCardHookExchangeAPDU(cardHandle, &transmitBuffer);
		if (rv!=SCARD_S_SUCCESS) {
#ifdef _DEBUG
			printf ("SCardTransmit Failed: %x: %s", rv, Stringify_Error(rv));
#endif
			free(buffer);
			return PCSC_To_PIV(rv);
		}
	}

	status = ReadStatusWord(&transmitBuffer);

	if (status!=PIV_SW_SUCCESS && (status&0xff00)!=PIV_SW_SUCCESS_DATA) {
		free(buffer);
		return status;
	}

	populateCache(&pDataCache, cardHandle, GENERATE_ASYMMETRIC_KEY_PAIR, &keyReference, 1, transmitBuffer.apduResponse, transmitBuffer.apduResponseSize - 2);

	GetAllResponses(cardHandle, status);
    
	/* if the return buffer is not NULL and the provided length is large enough, fill buffer and return success */
	if (publicKeyDataObjects != NULL) {
		if (pPublicKeyDataObjectsLength != NULL) {
			if (*pPublicKeyDataObjectsLength >= pDataCache->dataLen) {
				memcpy (publicKeyDataObjects, pDataCache->data, pDataCache->dataLen);
				*pPublicKeyDataObjectsLength = pDataCache->dataLen;
				free(buffer);
				return PIV_SW_SUCCESS;
			}
		}
	} 
	/* if you get here, the buffer is NULL or too small, indicate length, and return */
	
	*pPublicKeyDataObjectsLength = pDataCache->dataLen;
	free(buffer);
	return PIV_SW_SUCCESS;
}
PIV_RV PIV_Stringify_Response (PIV_ULong32 rv, PIV_UChar8 *errorBuffer, PIV_ULong32 *bufferLength) {
	switch(rv) {
		case PIV_OK:
		case PIV_CONNECTION_DESCRIPTION_MALFORMED:
		case PIV_CONNECTION_FAILURE:
		case PIV_CONNECTION_LOCKED:
//		case PIV_CARD_APPLICATION_NO_LONGER_AVAILABLE:
		case PIV_INVALID_CARD_HANDLE:
		case PIV_CARD_APPLICATION_NOT_FOUND:
		case PIV_AUTHENTICATOR_MALFORMED:
		case PIV_AUTHENTICATION_FAILURE:
		case PIV_INVALID_OID:
//		case PIV_DATA_OBJECT_NOT_FOUND:
		case PIV_SECURITY_CONDITIONS_NOT_SATISFIED:
//		case PIV_NO_CURRENT_CONTEXT:
		case PIV_INVALID_KEYREF_OR_ALGORITHM:
		case PIV_INSUFFICIENT_BUFFER:
		case PIV_INPUT_BYTES_MALFORMED:
		case PIV_INSUFFICIENT_CARD_RESOURCE:
//		case PIV_OFFSET_BEYOND_END_OF_DATA_CONTENT:
//		case PIV_GENERATED_KEY_PAIR_INCOMPATIBLE_WITH_EXISTING_KEY_PAIR:
		case PIV_UNSUPPORTED_CRYPTOGRAPHIC_MECHANISM:
			break;
	}
	return 0;
}
