// Card Manager
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include <Card_Manager.h>
#include "Card_Service.h"
#include <Java_Card_Manager.h>
#include <File_Manager.h>
#include <clains.h>
#include <sw.h>


Card_Manager::Card_Manager ( )
    {
      // create the JavaCard Manager
      managers[Registry_Entry::manager_javacard] = new Java_Card_Manager ;

      // create the Service Interface through which
      //   the File Manager will access the Card Manager
      Card_Service * card_service_for_javacard_manager = new Card_Service ;
      
      card_service_for_javacard_manager -> the_card_manager = this ;

      managers[Registry_Entry::manager_javacard] -> card_service
        = (Card_Service_Interface *) card_service_for_javacard_manager ;

      // create the File Manager
      managers[Registry_Entry::manager_file] =  new File_Manager ;

      // create the Service Interface through which
      //   the File Manager will access the Card Manager
      Card_Service * card_service_for_file_manager = new Card_Service ;
      
      card_service_for_file_manager -> the_card_manager = this ;

      managers[Registry_Entry::manager_file] -> card_service
        = (Card_Service_Interface *) card_service_for_file_manager ;

      managers[Registry_Entry::manager_invalid] = NULL;

      active_manager_index = Registry_Entry::manager_invalid;

      current_AID_length = 0;
    } ;


static unsigned char FileApplicationAID[] = { 0xA0, 0x00, 0x00, 0x01, 0x16, 0x3F, 0x00 } ;

long
Card_Manager::initialize ( int v )
  {
    if ( Card_Registry_Pointer ( ) == the_registry_pointer )
      the_registry_pointer = 
        Card_Registry_Pointer::object_pointer_to_indexed_pointer(new Card_Registry ) ;

    // initialize the File Manager
    managers[Registry_Entry::manager_file] -> initialize ( v ) ;

    // initialize the JavaCard Manager
    managers[Registry_Entry::manager_javacard] -> initialize ( v ) ;

    active_manager_index = Registry_Entry::manager_file ;

#ifdef PREBUILT_VM_EEPROM
    add_preloaded_applets_to_registry ( );
#endif

    return 0;
  }

inline
void get_length_value ( Octet_String the_payload , Memory::offset_t & offset , 
                        Octet & length , Octet_String & the_value)
  {
    Memory::offset_t the_start = the_payload . start ( );
    length = the_payload [ offset ] ;
    offset ++ ;
    the_value = Octet_String ( the_payload , the_start + offset , the_start + offset + length ) ;
    offset += length ;
  }

long
Card_Manager::install_for_load ( Octet_String the_payload )
  {
    Memory::offset_t offset = 0; // start parsing at beginning of data

    // parse the executable load file AID
    Octet load_file_AID_length ;
    Octet_String load_file_AID ; 
    get_length_value ( the_payload , offset , 
                        load_file_AID_length , 
                        load_file_AID ) ;

    // parse the executable load module AID
    Octet security_domain_AID_length ;
    Octet_String security_domain_AID ;
    get_length_value ( the_payload , offset , 
                        security_domain_AID_length , 
                        security_domain_AID ) ;

    // parse Load File Data Block Hash
    Octet load_file_data_block_hash_length ;
    Octet_String load_file_data_block_hash ; 
    get_length_value ( the_payload , offset , 
                        load_file_data_block_hash_length ,
                        load_file_data_block_hash ) ;

    // parse Load Parameters
    Octet load_parameters_length ;
    Octet_String load_parameters ; 
    get_length_value ( the_payload , offset , 
                        load_parameters_length ,
                        load_parameters ) ;

    // parse Load Token
    Octet load_token_length ;
    Octet_String load_token ; 
    get_length_value ( the_payload , offset , 
                        load_token_length ,
                        load_token ) ;

    // parse Format / Architecture Indicators
    Octet load_indicators_length ;
    Octet_String load_indicators ; 
      get_length_value ( the_payload , offset , 
                          load_indicators_length ,
                          load_indicators ) ;
    Octet format_indicator = LOAD_FILE_FORMAT_JAVACARD_CAP,
          architecture_indicator = TARGET_ARCHITECTURE_JAVA_CARD_BYTECODE;
    if ( 1 <= load_indicators_length )
      {
        format_indicator = load_indicators [ 0 ];
        if ( 2 <= load_indicators_length )
          architecture_indicator = load_indicators [ 1 ];
      }
    
    // TODO: search registry for referenced security domain.
    //       this check will also serve to enforce "only ISD" for now.

    // create registry entry for new load file
    if ( ! the_registry_pointer
             -> add_entry ( load_file_AID,
                            Registry_Entry::content_type_executable_load_file ) )
      return 1 ;

    // associate achitecture with new load file via manager
    Registry_Entry::Manager manager ;
    switch ( architecture_indicator )
      {
        case TARGET_ARCHITECTURE_JAVA_CARD_BYTECODE :
          manager = Registry_Entry::manager_javacard ;
          break ;
        case TARGET_ARCHITECTURE_7816_FILESYSTEM :
          manager = Registry_Entry::manager_file ;
          break ;
        default :
          manager = Registry_Entry::manager_invalid ;
          break ;
      }

    if ( ! the_registry_pointer
             -> set_manager ( load_file_AID,
                              Registry_Entry::content_type_executable_load_file,
                              manager ) )
      return 1 ;
    
    // TODO: record format of incoming load file

    // TODO: set resource allocation for new load file

    pending_load_file_AID = Octet_String::dup ( load_file_AID );
    return 0 ;
  } ;

long
Card_Manager::install_for_install_and_make_selectable ( Octet_String the_payload )
  {
    Memory::offset_t offset = 0; // start parsing at beginning of data
    long result = 0 ;

    // parse the executable load file AID
    Octet load_file_AID_length ;
    Octet_String load_file_AID ; 
    get_length_value ( the_payload , offset , 
                        load_file_AID_length ,
                        load_file_AID) ;

    // parse the executable load module AID
    Octet module_AID_length ;
    Octet_String module_AID ; 
    get_length_value ( the_payload , offset , 
                        module_AID_length ,
                        module_AID ) ;

    // parse the application AID
    Octet application_AID_length ;
    Octet_String application_AID ; 
    get_length_value ( the_payload , offset , 
                        application_AID_length ,
                        application_AID ) ;

    // parse the privileges
    Octet privileges_length ;
    Octet_String privileges ; 
    get_length_value ( the_payload , offset , 
                        privileges_length ,
                        privileges ) ;

    if ( 1 != privileges_length )
      return 1 ; // length of privileges must be only one byte

    // parse the install parameters
    Octet install_parameters_length ;
    Octet_String install_parameters ; 
    get_length_value ( the_payload , offset , 
                        install_parameters_length ,
                        install_parameters ) ;

    // parse the install token
    Octet install_token_length ;
    Octet_String install_token ; 
    get_length_value ( the_payload , offset , 
                        install_token_length ,
                        install_token ) ;

    // read manager from registry entry for load file
    Registry_Entry::Manager manager =
      the_registry_pointer -> get_manager ( load_file_AID,
                                  Registry_Entry::content_type_executable_load_file );
    
    // invoke the load method of the appropriate Application Manager
    if ( Registry_Entry::manager_invalid == manager )
      return 1 ;

    Load_Module * the_load_module = 
      the_registry_pointer -> get_data_blocks ( load_file_AID,
                                      Registry_Entry::content_type_executable_load_file ) ;

    // TODO: what if get_data_blocks returns NULL?
    if ( 0 == the_load_module )
      return 1 ;

    Application_Manager & the_manager = * managers [manager] ;

    result = the_manager . store_load_module ( 0 , 
                                              application_AID, 
                                              * the_load_module ) ;
    if ( 0 != result )
      return result ;
    
    Octet_String runnable_instance_AID ( 16 ) ; // Max AID length is 16
    Octet_String::cpy ( runnable_instance_AID , application_AID ) ;
    result = the_manager . create_runnable_instance( 0 , 
                                                    module_AID, 
                                                    application_AID ,
                                                    install_parameters ,
                                                    runnable_instance_AID ) ;
    if ( 0 != result )
      return result ;

    // record the new application
    if ( ! the_registry_pointer -> add_entry ( runnable_instance_AID ,
                                      Registry_Entry::content_type_application ) )
      return 1 ;
    
    // note the associated manager
    if ( ! the_registry_pointer -> set_manager ( runnable_instance_AID ,
                                        Registry_Entry::content_type_application ,
                                        manager ) )
      return 1 ;

    return 0 ;
  }

long
Card_Manager::install_for_make_selectable ( Octet_String the_payload )
  {
    Memory::offset_t offset = 2; // skip the two mandatory zeros
    long result = 0 ;

    // parse the application AID
    Octet application_AID_length ;
    Octet_String application_AID ; 
    get_length_value ( the_payload , offset , 
                        application_AID_length ,
                        application_AID ) ;

    // parse the privileges
    Octet privileges_length ;
    Octet_String privileges ; 
    get_length_value ( the_payload , offset , 
                        privileges_length ,
                        privileges ) ;

    if ( 1 != privileges_length )
      return 1 ; // length of privileges must be only one byte

    // parse the install parameters
    Octet install_parameters_length ;
    Octet_String install_parameters ; 
    get_length_value ( the_payload , offset , 
                        install_parameters_length ,
                        install_parameters ) ;

    // parse the install token
    Octet install_token_length ;
    Octet_String install_token ; 
    get_length_value ( the_payload , offset , 
                        install_token_length ,
                        install_token ) ;

    // use the default manager
    Registry_Entry::Manager manager =
      the_registry_pointer -> default_manager ( );
    
    // invoke the load method of the appropriate Application Manager
    if ( Registry_Entry::manager_invalid == manager )
      return 1 ;

    Application_Manager & the_manager = * managers [manager] ;

    Octet_String runnable_instance_AID ( 16 ) ;
    Octet_String::cpy ( runnable_instance_AID , application_AID ) ;

    result = the_manager . create_runnable_instance( 0 , 
                                                    application_AID , 
                                                    application_AID ,
                                                    install_parameters, 
                                                    runnable_instance_AID );
    if ( 0 != result )
      return result ;

    // record the new applicaton in the registry
    if ( ! the_registry_pointer -> add_entry ( runnable_instance_AID ,
                                      Registry_Entry::content_type_application ) )
      return 1 ;
    
    // note the associated manager
    if ( ! the_registry_pointer -> set_manager ( runnable_instance_AID ,
                                        Registry_Entry::content_type_application ,
                                        manager ) )
      return 1 ;

    return 0 ;
  }

long
Card_Manager::exchange_APDUs ( const CAPDU_Packet & command, RAPDU_Packet & response )
  {  
    long result = 0 ;
    response . set_APDU_length ( 0 ) ;

    command . log ( "Card_Manager::exchange_APDUs:CAPDU_Packet: " ) ;

    // is the ISD the current application?
    bool current_app_is_ISD = ( current_AID_length == ISD_AID_length 
                                && 0 == memcmp ( current_AID , ISD_AID , current_AID_length ) ) ;

    // allow all APDUs to pass through the ISD handler filter for now.
    // TODO: remove this crutch when proper ISD is part of configuration
    current_app_is_ISD = true;

    const Octet_String the_payload ( command . payload ( ) , command . get_P3 ( ) ) ;
    
    // ISD APDU Handlers
    if ( current_app_is_ISD )
      switch ( command . get_CLAINS ( ) )
        {
          case CAPDU_Packet::CLAINS_INSTALL:
            {
              switch ( command . get_P1 ( ) )
                {
                  case P1_INSTALL_FOR_LOAD:
                    result = install_for_load ( the_payload ) ;
                    if ( 0 == result )
                      goto return_SW_OK;
                    goto done ;
                    // do not fall through to other handlers.

                  case P1_INSTALL_FOR_INSTALL_AND_MAKE_SELECTABLE:
                    result = install_for_install_and_make_selectable ( the_payload ) ;
                    if ( 0 == result )
                      goto return_SW_OK;
                    goto done ;
                    // do not fall through to other handlers.

                  case P1_INSTALL_FOR_MAKE_SELECTABLE:
                    result = install_for_make_selectable ( the_payload ) ;
                    if ( 0 == result )
                      goto return_SW_OK;
                    goto done ;
                    // do not fall through to other handlers.

                  default:
                    break;
                } // switch ( command . get_P1 ( ) )
              break;
            }

          case CAPDU_Packet::CLAINS_LOAD:
            {
              Memory::offset_t offset = 0; // start parsing at beginning of data
              // TODO: process DAP blocks

              // parse load file data block TLV
              if ( the_payload [ offset ] != TAG_LOAD_FILE_DATA_BLOCK )
                {
                  result = 1;
                  goto done ;
                } // bad command data

              offset++;

              Octet data_block_length ;
              Octet_String data_block ;
              get_length_value ( the_payload , offset , 
                                  data_block_length , 
                                  data_block ) ;

              // store the load file data block in the list
              the_registry_pointer -> store_data_block ( pending_load_file_AID ,
                                                Registry_Entry::content_type_executable_load_file,
                                                data_block );

              goto return_SW_OK; // do not fall through to other handlers.
            }
          
          default:
            break; // TODO: when ISD selection is fixed, turn this into an error.  Don't know this APDU.
        } // switch ( command . get_CLAINS ( ) )
    // if ( current_app_is_ISD )
    

    if ( CAPDU_Packet::CLAINS_SELECT == command . get_CLAINS ( ) 
         && 0x04 == command . get_P1 ( ) )  // select by AID
      {
        // parse the AID in this APDU
        Octet_String AID ( command . payload ( ) , command . get_P3 ( ) ) ;

        // map truncated AIDs to untruncated AIDs
        // TODO: make this general
        if ( AID == Octet_String ( truncated_PIV_AID, sizeof ( truncated_PIV_AID ) ) )
        {
            AID = Octet_String ( full_V_01_PIV_AID , sizeof (  full_V_01_PIV_AID ) ) ;
        }

        // fetch the manager for the app with this AID in registry
        Registry_Entry::Manager manager
          = the_registry_pointer -> get_manager ( AID , Registry_Entry::content_type_application ) ;
        
        if ( Registry_Entry::manager_invalid != manager )
          {
            active_manager_index = manager;
            // remember the current AID
            Octet_String::cpy ( current_AID , sizeof ( current_AID ) , AID ) ;
            current_AID_length =  ( unsigned int ) AID . length ( );          
          }
        else 
          {
            response . set_status ( RAPDU_Packet::SW_APP_NOT_FOUND ) ;
            goto done ;
          }
      }

    // dispatch to the manager associated with the current app
    if ( Registry_Entry::manager_invalid != active_manager_index 
         && NULL != managers [ active_manager_index ] )
      result = managers [active_manager_index] -> exchange_APDUs ( command, response ) ;
    else
      result = 1 ;
    goto done ;

  return_SW_OK:
    response . set_status ( RAPDU_Packet::SW_OK ) ;
    goto done ;

  done:
    response . log ( "Card_Manager::exchange_APDUs:RAPDU_Packet: " ) ;
    return result ;
  }


long
Card_Manager::finalize ( )
  {

    for ( int i = 0 ; i < Registry_Entry::manager_invalid ; i++ )
    {
      if ( managers[i] )
      {
        // TODO: each application manager should clean up its own card service
        if ( managers[i] -> card_service )
        {
          delete managers[i] -> card_service ;
        }
        
        managers[i] -> finalize ( ) ;

        delete managers[i] ;
      }
    }
    
    return 0;
  }


long
Card_Manager::power_up ( void )
  {
    long return_temp, return_code = 0 ; 

    for ( int i = 0 ; i < Registry_Entry::manager_invalid ; i++ )
      if ( managers[i] )
        if ( return_temp = managers[i] -> power_up ( ) )
          return_code = return_temp ;

    return return_code ;  
  }


long
Card_Manager::power_down ( void )
  {
    long return_temp, return_code = 0 ;

    for ( int i = 0 ; i < Registry_Entry::manager_invalid ; i++ )
    {
      if ( managers[i] )
      {
        if ( return_temp = managers[i] -> power_down ( ) )
        {
          return_code = return_temp ;
        }
      }
    }

    return return_code ;
  }


void 
Card_Manager::get_ATR ( unsigned char * ATR , 
                        unsigned int & ATR_length )
  {
    return managers [ active_manager_index ] -> get_ATR ( ATR , ATR_length ) ;
  }


unsigned char
Card_Manager::get_card_content_state ( Card_Service * caller )
  {
    return 0 ;
  }


bool
Card_Manager::set_card_content_state ( Card_Service * caller,
                                       unsigned char new_state )
  {
    return false;
  }

  
unsigned char
Card_Manager::get_card_state( Card_Service * caller )
  {
    return 0;
  }

  
bool
Card_Manager::lock_card ( Card_Service * caller )
  {
    return false;
  }
  

bool
Card_Manager::terminate_card ( Card_Service * caller )
  {
    return false;
  }


bool
Card_Manager::store_persistent_state ( FILE * the_persistent_storage_output_file , unsigned short & checksum )
  {
    bool result = the_registry_pointer . store ( the_persistent_storage_output_file , checksum ) ;

    for ( int i = 0 ; i < Registry_Entry::manager_invalid ; i++ )
      if ( managers[i] )
        result &= managers[i] -> store_persistent_state ( the_persistent_storage_output_file , checksum ) ;
    
    return result ;
  } ;


bool
Card_Manager::load_persistent_state ( FILE * the_persistent_storage_input_file , unsigned short & checksum )
  {
    bool result = the_registry_pointer . load ( the_persistent_storage_input_file , checksum ) ;

    for ( int i = 0 ; i < Registry_Entry::manager_invalid ; i++ )
      if ( managers[i] )
        result &= managers[i] -> load_persistent_state ( the_persistent_storage_input_file , checksum ) ;

    return result ;
  } ;

void
Card_Manager::add_preloaded_applets_to_registry( void )
  {
    return;
  }

bool
Card_Manager::set_ATR_hist_bytes ( Card_Service * caller,
                                   Octet_String & new_hist_bytes )
  {
    return false;
  }
