// File System Application
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if defined(_WIN32) && defined(_DEBUG)
#include <windows.h>
#endif
#include <File_System_Application.h>
#include <macros.h>

long File_System_Application::read_binary ( const CAPDU_Packet & command, RAPDU_Packet & response )
  {
    the_file_system_pointer -> verify_current_elementary_file ( ) ;
    the_file_system_pointer -> verify_elementary_file_authorization ( ACR_AM_FILE_READ ) ;

    unsigned short offset = command . get_P1P2 ( ) ;
    Octet length = command . get_P3 ( ) ;
    Octet_String data ( length ) ;
    the_file_system_pointer -> read_binary ( data , offset , length ) ;
    APDU ( data ) . store_into_APDU_packet ( response ) ;
    return 0 ;
  } ;

long File_System_Application::select ( const CAPDU_Packet & command, RAPDU_Packet & response )
  {
    file_identifier_t file_identifier ;
    
    const Octet P1 = command . get_P1 ( ) ;
    const Octet P2 = command . get_P2 ( ) ;
    const Octet P3 = command . get_P3 ( ) ;

    switch ( P1 )
      {
        case 0x00:
        case 0x01:
        case 0x02:
          if ( 2 == P3 )
            file_identifier = command . get_payload_u16 ( ) ;
          else
            throw RAPDU_Packet::SW_LC_DOES_NOT_MATCH_P1_P2 ;
          break ;

        case 0x03:
          if ( 0 == P3 )
            file_identifier = PARENT_DIR ;
          else if ( 2 == P3 )
            {
              file_identifier = command . get_payload_u16 ( ) ;
              if ( MF != file_identifier )
                throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
            }
          else
            throw RAPDU_Packet::SW_LC_DOES_NOT_MATCH_P1_P2 ;
          break ;

        case 0x04:
          if ( the_file_manager_pointer 
                 -> is_known_AID ( Octet_String ( command . payload ( ) , P3 )))
            throw RAPDU_Packet::SW_OK ;
          else
            throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
          break ;

        default:
          throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
      } ;

    switch ( P2 )
      {
        case 0x00:
        case SELECT_NO_DATA:
          break ;

        default:
          throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
      } ;

    the_file_system_pointer -> select_file ( file_identifier , P1 ) ;  

    if ( 0x00 == P2 )
      {
        Octet_String the_FCP_octet_string =
          the_file_system_pointer 
            -> get_current_file_pointer ( ) 
                -> FCP_octet_string ;
        APDU ( the_FCP_octet_string ) . store_into_APDU_packet ( response ) ;
      }

    return 0 ;
  } ;

long File_System_Application::update_binary ( const CAPDU_Packet & command, RAPDU_Packet & response )
  {
    the_file_system_pointer -> verify_current_elementary_file ( ) ;
    the_file_system_pointer -> verify_elementary_file_authorization ( ACR_AM_FILE_UPDATE ) ;

    Octet_String data ( command . payload ( ) , command . get_P3 ( ) ) ;
    the_file_system_pointer -> update_binary ( data , command . get_P1P2 ( ) ) ;
    return 0 ;
  } ;

long File_System_Application::get_challenge ( const CAPDU_Packet & command, RAPDU_Packet & response )
  {
    if ( 0 != command . get_P1P2 ( ) )
      throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;

    Memory::size_t length = command . get_P3 ( ) ;
    if ( 0 < length )
      {
        Octet_String & challenge = * new Octet_String ( length ) ;
        the_security_system_pointer -> generate_cryptogram ( challenge , length ) ;
        the_challenge_pointer = 
          Octet_String_Pointer::object_pointer_to_indexed_pointer ( & challenge ) ;
        APDU ( challenge ) . store_into_APDU_packet ( response ) ;
      }
    return 0 ;
  } ;

long File_System_Application::external_authenticate ( const CAPDU_Packet & command, RAPDU_Packet & response )
  {

    Octet the_algorithm_identifier = command . get_P1 ( ) ;

    Octet the_key_number = command . get_P2 ( ) ; // key number -- 0x00 => default

    the_security_system_pointer
      -> set_authentication ( the_key_number , false ) ;

    if ( Octet_String_Pointer ( ) == the_challenge_pointer )
      throw RAPDU_Packet::SW_CONDITIONS_OF_USE_NOT_SATISFIED ;

    Memory::size_t the_challenge_length = the_challenge_pointer -> length ( ) ;

    if ( command .get_P3 ( ) != the_challenge_length )
      throw RAPDU_Packet::SW_CONDITIONS_OF_USE_NOT_SATISFIED ; // SW_WRONG_LENGTH?

    Octet_String the_cryptogram ( command . payload ( ) , command . get_P3 ( ) ) ;

    Octet_String the_plaintext  ( the_challenge_length ) ;

    the_security_system_pointer 
      -> decrypt_symmetric ( command . get_P1 ( ) , // Algorithm identifier
                            the_key_number ,
                            the_cryptogram ,       // cryptogram
                            the_plaintext ) ;      // plaintext

    bool authenticated = 
      0 == Octet_String::cmp ( * the_challenge_pointer , 
                              the_plaintext , 
                              the_challenge_length ) ;

    the_security_system_pointer
      -> set_authentication ( the_key_number , authenticated ) ;

    if ( ! authenticated )
      throw RAPDU_Packet::SW_VERIFICATION_FAILED ;

    return 0 ;
  } ;


long File_System_Application::internal_authenticate ( const CAPDU_Packet & command, RAPDU_Packet & response )
  {
    Memory::size_t the_challenge_length = command .get_P3 ( ) ;

    // Omit the following test to allow challenge lengths other than 8
    if ( 8 != the_challenge_length )
      throw RAPDU_Packet::SW_CONDITIONS_OF_USE_NOT_SATISFIED ; // SW_WRONG_LENGTH?

    Octet_String the_plaintext  ( command . payload ( ) , command . get_P3 ( ) ) ;

    Octet_String the_cryptogram ( the_challenge_length ) ;

    the_security_system_pointer 
      -> encrypt_symmetric ( command . get_P1 ( ) , // Algorithm identifier
                            command . get_P2 ( ) , // key number -- 0x00 => default
                            the_plaintext ,        // plaintext
                            the_cryptogram ) ;     // cyphertext

    APDU ( the_cryptogram ) . store_into_APDU_packet ( response ) ;

    return 0 ;
  } ;


long File_System_Application::verify_pin ( const CAPDU_Packet & command, RAPDU_Packet & response )
  {
    if ( 0 != command . get_P1 ( ) )
      throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
    
    Key_Reference the_PIN_reference = command . get_P2 ( ) ;

    Octet_String the_presented_PIN  ( command . payload ( ) , command . get_P3 ( ) ) ;

    the_security_system_pointer 
        -> verify_pin ( the_PIN_reference , the_presented_PIN ) ;
    return 0 ;
  } ;

long File_System_Application::manage_security_environment ( const CAPDU_Packet & command, RAPDU_Packet & response )
  {

    if ( ! (  MSE_SET_DST == command . get_P1 ( ) 
            && CRT_DIGITAL_SIGNATURE == command . get_P2 ( ) ) )
      throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
    
    // GSC-ISv2.1 says payload is Key Reference -- where is algorithm specified?
    
    Key_Reference_Information 
      key_reference_information ( command . payload ( ) , command . get_P3 ( ) ) ;

    the_security_system_pointer -> the_security_environment . the_key_reference =
      key_reference_information . key_reference ( ) ;
    
    return 0 ;
  } ;

long File_System_Application::perform_security_operation ( const CAPDU_Packet & command, RAPDU_Packet & response )
  {
    if ( ! (  PSO_DIGITAL_SIGNATURE == command . get_P1 ( ) 
            && PSO_DIGITAL_SIGNATURE_DATA == command . get_P2 ( ) ) )
      throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
    
    Memory::size_t length = command . get_P3 ( ) ;

    // Since MSE cannot set the algorithm, PSO can assume it is RSA_1024
    // Hence the cryptogram is 1024 bits = 128 Octets.

    if ( 0 == length || 128 < length )
      throw RAPDU_Packet::SW_WRONG_LENGTH ;

    Octet_String cryptogram ( 128 ) ;

    the_security_system_pointer 
      -> encrypt ( Octet_String ( command . payload ( ) , length ) , cryptogram ) ;

    response . set_APDU_length ( 128 ) ;

    Octet_String::cpy ( response . payload ( ) , 128 , cryptogram ) ;

    return 0 ;                
  } ;


File_System_Application::Dispatch_Table_Entry File_System_Application::dispatch_table [ ] =
  {
    { CAPDU_Packet::CLAINS_READ_BINARY                 , read_binary                 } ,
    { CAPDU_Packet::CLAINS_SELECT                      , select                      } ,
    { CAPDU_Packet::CLAINS_UPDATE_BINARY               , update_binary               } ,
    { CAPDU_Packet::CLAINS_EXTERNAL_AUTHENTICATE       , external_authenticate       } ,
    { CAPDU_Packet::CLAINS_GET_CHALLENGE               , get_challenge               } ,
    { CAPDU_Packet::CLAINS_INTERNAL_AUTHENTICATE       , internal_authenticate       } ,
    { CAPDU_Packet::CLAINS_VERIFY                      , verify_pin                  } ,
    { CAPDU_Packet::CLAINS_MANAGE_SECURITY_ENVIRONMENT , manage_security_environment } ,
    { CAPDU_Packet::CLAINS_PERFORM_SECURITY_OPERATION  , perform_security_operation  } 
  } ; 


long File_System_Application::exchange_APDUs ( const CAPDU_Packet & command , 
                                               RAPDU_Packet & response )
  {
    // Dispatch on the INS 
    const CLAINS clains = command . get_CLAINS ( );
    const int dispatch_table_size = sizeof ( dispatch_table ) / sizeof ( dispatch_table [ 0 ] ) ;
    for ( int i = 0  ; i < dispatch_table_size ; i ++ )
      if ( dispatch_table [ i ] . clains == clains )
        break ;

    if ( i == dispatch_table_size )
      throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;

    return ( this ->* dispatch_table [ i ] . APDU_handler ) ( command, response ) ;
  } ;



#define FILE_SYSTEM_APP_AID GSA_RID , NETWORK_WORD_BYTES ( MF ) 

static Octet the_file_system_app_AID_octets [ ] = { FILE_SYSTEM_APP_AID } ;

File_System_Application::File_System_Application( File_Manager * file_manager_pointer ,
                                                  const File_System_Pointer & file_system_pointer ,
                                                  const Security_System_Pointer & security_system_pointer ,
                                                  Octet_String_Pointer & challenge_pointer )
  : Application( file_manager_pointer ,
                 file_system_pointer ,
                 security_system_pointer ,
                 challenge_pointer ) ,
    the_AID ( the_file_system_app_AID_octets , sizeof ( the_file_system_app_AID_octets ) ) 
  {
    const int dispatch_table_size = sizeof ( dispatch_table ) / sizeof ( dispatch_table [ 0 ] ) ;
    for ( int i = 0  ; i < dispatch_table_size ; i ++ )
      file_manager_pointer -> record_APDU_handler ( dispatch_table [ i ] . clains , this ) ;
  } ;

FILE_SYSTEM_APPLICATION_DLL Application * 
create_application ( File_Manager * file_manager_pointer ,
                     const File_System_Pointer & file_system_pointer ,
                     const Security_System_Pointer & security_system_pointer ,
                     Octet_String_Pointer & challenge_pointer  )
  {
    return 
      new File_System_Application ( file_manager_pointer ,
                                    file_system_pointer ,
                                    security_system_pointer ,
                                    challenge_pointer ) ;
  } ;
