// File System
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if !defined(__FILE_SYSTEM_H_INCLUDED__)
#define __FILE_SYSTEM_H_INCLUDED__

#if defined(_MSC_VER )
  #pragma warning ( disable : 4702 )  // unreachable code in <list>
  #if defined(FILE_SYSTEM_EXPORTS)
    #define FILE_SYSTEM_DLL _declspec(dllexport)
  #else //  defined ( FILE_SYSTEM_EXPORTS )
    #define FILE_SYSTEM_DLL _declspec(dllimport)
  #endif  // defined ( FILE_SYSTEM_EXPORTS )
#else
   #define FILE_SYSTEM_DLL
#endif

#include <TLV.h>
#include <list>
#include <iostream>

#include <Messages.h>

#include <Octet_String.h>

#include <macros.h>

#include <APDU_Packet.h>

// Some object has to be the root object
// That will be the file system, so it has
// to keep a pointer to the other candidate,
// the security system.
#include <Security_System.h>

/* ISO/IEC 7816-4 Standard File Identifiers */
#define MF              0x3F00
#define EF_DIR          0x2F00      /* SFI 0x1E */
#define EF_ATR          0x2F01
#define EF_PL           0x2F05      /* SFI 0x05 */
#define EF_ARR          0x2F06      /* SFI 0x06 */
#define EF_ICCID        0x2FE2      /* SFI 0x02 */

// GSC-IS Object Identifiers
#define DO_PIV_APP_X509                 0x0500
#define DO_CHUID                        0x3000
#define DO_FINGER_1                     0x6010
#define DO_FINGER_2                     0x6011
#define DO_LOGICAL_X509                 0x0101
#define DO_SECURITY_OBJECT              0x9000
#define DO_CCC                          0xDB00
#define DO_FACIAL_IMAGE                 0x6030
#define DO_PRINTED_INFO                 0x3001
#define DO_SIGNING_X509                 0x0100
#define DO_KEY_MANAGEMENT_X509          0x0102

// SP 800-73 BER-TLV Tags
#define TAG_PIV_APP_X509     0x5FC101 // 0x0500
#define TAG_CHUID            0x5FC102 // 0x3000
#define TAG_FINGER_1         0x5FC103 // 0x6010
#define TAG_FINGER_2         0x5FC104 // 0x6011
#define TAG_LOGICAL_X509     0x5FC105 // 0x0101
#define TAG_SECURITY_HASH    0x5FC106 // 0x9000
#define TAG_CCC              0x5FC107 // 0xDB00
#define TAG_FACIAL_IMAGE     0x5FC108 // 0x6030
#define TAG_PRINTED_INFO     0x5FC109 // 0x3001
#define TAG_SIGNING_X509     0x5FC10A // 0x0100
#define TAG_KEY_X509         0x5FC10B // 0x0102

// SP 800-73 Data Object Sizes (in bytes)
/*
#define DO_PIV_X509_SIZE                 1856
#define DO_CHUID_SIZE                    3377
#define DO_FINGER_1_SIZE                 7768
#define DO_FINGER_2_SIZE                 7768
#define DO_CARD_X509_SIZE                1856
#define DO_SECURITY_OBJECT_SIZE          1000
#define DO_CCC_SIZE                      266
#define DO_FACIAL_IMAGE_SIZE             12704
#define DO_PRINTED_INFO_SIZE             106
#define DO_SIGNING_X509_SIZE             1856
#define DO_KEY_MANAGEMENT_X509_SIZE      1856
*/
//Per Bill MacGregor - sizes will be removed from spec. Increaze ~ 20% for now. Used length from Conf Testing
#define DO_PIV_X509_SIZE                 1905 //1856
#define DO_CHUID_SIZE                    3393 //3377
#define DO_FINGER_1_SIZE                 7768 //7768
#define DO_FINGER_2_SIZE                 7768 //7768
#define DO_CARD_X509_SIZE                1905 //1856
#define DO_SECURITY_OBJECT_SIZE          1008 //1000
#define DO_CCC_SIZE                      297 //266
#define DO_FACIAL_IMAGE_SIZE             12710 //12704
#define DO_PRINTED_INFO_SIZE             120 //106
#define DO_SIGNING_X509_SIZE             1905 //1856
#define DO_KEY_MANAGEMENT_X509_SIZE      1905 //1856

// ISO/IEC 7816-15 Symmetric_Keystore DFs and EFs

/* Relative and Special File Identifiers */
#define CURRENT_EF      0x1FFF
#define CURRENT_DIR     0x3FFF
#define PARENT_DIR      0x5FFF
#define CURRENT_ADF     0x7FFF

/* SelectFile options */
#define SELECT_NO_DATA 0x0C

/* READ RECORDS and UPDATE RECORD */
#define NEXT_RECORD           0x02
#define PREVIOUS_RECORD       0x03
#define ABSOLUTE_RECORD       0x04  // also current record, if P1=0

typedef unsigned short file_identifier_t ;
typedef unsigned int   data_object_tag_t ;
typedef unsigned short object_identifier_t ;

class FILE_SYSTEM_DLL Malformed_File_Identifier_Exception { } ;

class FILE_SYSTEM_DLL File_Identifier : public Primitive_BER_TLV
{
public:
  File_Identifier ( const Primitive_BER_TLV & primitive_BER_TLV )
    : Primitive_BER_TLV ( primitive_BER_TLV )
    {
      if ( ! ( BER_TLV_Tag ( FCP_FID ) == primitive_BER_TLV . tag ( )
                &&  2 == primitive_BER_TLV . length ( ) . value ( ) ) )
        throw Malformed_File_Identifier_Exception ( ) ;        
    } ;

  void put_file_identifier ( const file_identifier_t file_identifier )
  {
    values ( ) . put_u16 ( 0, file_identifier ) ;
  } ;

  unsigned short get_file_identifier ( ) const
  {
    return values ( ) . get_u16 ( 0 ) ;
  } ;
} ;

class FILE_SYSTEM_DLL Lifecycle_Status : public Primitive_BER_TLV
  {
  public:

    Lifecycle_Status ( const Primitive_BER_TLV & the_LCS )
      : Primitive_BER_TLV ( the_LCS )
      { 
        if ( tag ( ) != BER_TLV_Tag ( FCP_LIFECYCLE ) )
          throw RAPDU_Packet::SW_INCOMPATIBLE_FILE_STRUCTURE ;
      } ;

    Octet 
    get_status ( ) const 
      { return the_values . get_octet ( 0 ) ; } ;

    Octet 
    set_status ( Octet new_status )  
      { return the_values . put_octet ( 0 , new_status ) ; } ;

  };
typedef indexed_pointer < Lifecycle_Status > Lifecycle_Status_Pointer ;


class FILE_SYSTEM_DLL File_Control_Parameters : public Constructed_BER_TLV
  {
  public:
    File_Control_Parameters ( const Octet * FCP_octets , size_t FCP_octets_size )
      : Constructed_BER_TLV ( Octet_String ( FCP_octets , FCP_octets_size ) )
      { } ;

    File_Control_Parameters ( const Octet_String & FCP_octet_string )
      : Constructed_BER_TLV ( FCP_octet_string )
      { } ;

    file_identifier_t file_identifier ( )
      {
        File_Identifier file_identifier_BER_TLV = 
          * Primitive_BER_TLV_Pointer ( find ( FCP_FID ) ) ;
        return file_identifier_BER_TLV . get_file_identifier ( ) ;
      }

    data_object_tag_t data_object_tag ( )
      {
        Constructed_BER_TLV_Pointer data_object_template_BER_TLV = 
            find ( FCP_DATA_OBJECT_TEMPLATE ) ;
        Primitive_BER_TLV_Pointer tag_list = data_object_template_BER_TLV -> find( FCP_TAG_LIST );
        return tag_list->values().get_u24( 0 ) ;
      }

    bool denotes_a_dedicated_file  ( )
      {
        Primitive_BER_TLV_Pointer fdb ( find ( FCP_FDB ) );
        return ( fdb->values ( )[ 0 ] == FDB_DF) || ( fdb->values ( )[ 0 ] == FDB_SHAREABLE_DF);
      } ;

    bool denotes_a_data_object  ( )
      {
        Primitive_BER_TLV_Pointer fdb ( find ( FCP_FDB ) );
        return ( fdb->values ( )[ 0 ] == FDB_BER_TLV_STRUCTURE ) || ( fdb->values ( )[ 0 ] == FDB_SHAREABLE_BER_TLV );
      } ;

    Memory::size_t file_allocation ( ) const
    {
      Primitive_BER_TLV_Pointer fdb = find ( FCP_FILEALLOC );
      return fdb == Primitive_BER_TLV_Pointer ( ) ? 0 : fdb -> values ( ) . get_i16 ( 0 ) ;
    };


    const Security_Attribute_Pointer security_attribute_pointer ( ) const 
      { 
        BER_TLV_Pointer the_ACR_pointer = find ( FCP_ACR_EXPANDED ) ; 

        if ( BER_TLV_Pointer ( ) != the_ACR_pointer )
          return Security_Attribute_Pointer::object_pointer_to_indexed_pointer
                   ( new Security_Attribute ( * Constructed_BER_TLV_Pointer ( the_ACR_pointer ) ) ) ;
        else
          return Security_Attribute_Pointer ( ) ;
      } ;

    const Primitive_BER_TLV_Pointer tag_list_pointer ( ) const 
      { 
        Constructed_BER_TLV_Pointer data_object_template_BER_TLV = find ( FCP_DATA_OBJECT_TEMPLATE ) ; 
    
        return data_object_template_BER_TLV -> find( FCP_TAG_LIST );
      } ;

    const Lifecycle_Status_Pointer lifecycle_status_pointer ( ) const 
      { 
        BER_TLV_Pointer the_LCS_pointer = find ( FCP_LIFECYCLE ) ; 

        if ( BER_TLV_Pointer ( ) != the_LCS_pointer )
          return Lifecycle_Status_Pointer::object_pointer_to_indexed_pointer
                   ( new Lifecycle_Status ( * Primitive_BER_TLV_Pointer ( the_LCS_pointer ) ) ) ;
        else
          return Lifecycle_Status_Pointer ( ) ;
      } ;

    Lifecycle_Status_Pointer lifecycle_status_pointer ( ) 
      { 
        BER_TLV_Pointer the_LCS_pointer = find ( FCP_LIFECYCLE ) ; 

        if ( BER_TLV_Pointer ( ) != the_LCS_pointer )
          return Lifecycle_Status_Pointer::object_pointer_to_indexed_pointer
                   ( new Lifecycle_Status ( * Primitive_BER_TLV_Pointer ( the_LCS_pointer ) ) ) ;
        else
          return Lifecycle_Status_Pointer ( ) ;
      } ;

  } ;

class FILE_SYSTEM_DLL Data_Object ;
typedef indexed_pointer < Data_Object > Data_Object_Pointer ;

class FILE_SYSTEM_DLL Dedicated_File ;
typedef indexed_pointer < Dedicated_File > Dedicated_File_Pointer ;

class FILE_SYSTEM_DLL Elementary_File ;
typedef indexed_pointer < Elementary_File > Elementary_File_Pointer ;

class FILE_SYSTEM_DLL File_System ;
typedef indexed_pointer < File_System > File_System_Pointer ;

class FILE_SYSTEM_DLL File : public Object
  {
  public:

    typedef Octet Type ;
    static const Type none        = 0 ;
    static const Type dedicated   = 1 ;
    static const Type elementary  = 2 ;
    static const Type data_object = 3 ;

    static const int dump_indent_spaces = 5 ;

    File ( const Dedicated_File_Pointer & parent_dedicated_file_pointer , 
           const Octet * FCP_octets , size_t FCP_octets_size )
      : the_parent_dedicated_file_pointer ( parent_dedicated_file_pointer ) ,
        FCP_octet_string ( FCP_octets , FCP_octets_size  )
      { } ;

    File ( const Dedicated_File_Pointer & parent_dedicated_file_pointer , 
           const Octet_String & FCP_octet_string )
      : the_parent_dedicated_file_pointer ( parent_dedicated_file_pointer ) ,
        FCP_octet_string ( FCP_octet_string )
      { } ;

    bool is_dedicated_file ( )
      { return get_FCP ( ) . denotes_a_dedicated_file ( ) ; } ;

    Dedicated_File_Pointer the_parent_dedicated_file_pointer ;

    File_Control_Parameters get_FCP ( ) 
      { return File_Control_Parameters ( FCP_octet_string ) ; } ;

    Octet_String FCP_octet_string ;
  } ;


class FILE_SYSTEM_DLL Data_Object: public File
  {
  public:
    typedef Data_Object_Pointer Pointer ;

    Data_Object ( const Dedicated_File_Pointer & parent_dedicated_file_pointer , 
                      const Octet * FCP_octets , 
                      Memory::size_t FCP_octets_size , 
                      Memory::size_t initial_data_size )
      : File ( parent_dedicated_file_pointer , FCP_octets , FCP_octets_size ) , 
        the_data ( Octets::alloc ( initial_data_size ) ) 
      { } ;

    Data_Object ( const Dedicated_File_Pointer & parent_dedicated_file_pointer , 
                      const Octet_String & FCP_octet_string , 
                      Memory::size_t initial_data_size )
      : File ( parent_dedicated_file_pointer , FCP_octet_string ) , 
        the_data (  Octets::alloc ( initial_data_size ) ) 
      { } ;

    void dump ( File_System * the_file_system_pointer , unsigned int indentation ) ;

    Octets::Pointer the_data ;
  } ;

class FILE_SYSTEM_DLL Elementary_File : public File
  {
  public:
    typedef Elementary_File_Pointer Pointer ;

    Elementary_File ( const Dedicated_File_Pointer & parent_dedicated_file_pointer , 
                      const Octet * FCP_octets , 
                      Memory::size_t FCP_octets_size , 
                      Memory::size_t initial_data_size )
      : File ( parent_dedicated_file_pointer , FCP_octets , FCP_octets_size ) , 
        the_data ( Octets::alloc ( initial_data_size ) ) 
      { } ;

    Elementary_File ( const Dedicated_File_Pointer & parent_dedicated_file_pointer , 
                      const Octet_String & FCP_octet_string , 
                      Memory::size_t initial_data_size )
      : File ( parent_dedicated_file_pointer , FCP_octet_string ) , 
        the_data (  Octets::alloc ( initial_data_size ) ) 
      { } ;

    void dump ( File_System * the_file_system_pointer , unsigned int indentation ) ;

    Octets::Pointer the_data ;
  } ;


class FILE_SYSTEM_DLL Dedicated_File : public File
  {
  public:
    
    typedef list < Dedicated_File_Pointer  , 
                   persistent_allocator < Dedicated_File_Pointer  > > 
      Dedicated_File_Pointers  ;

    typedef list < Elementary_File_Pointer , 
                   persistent_allocator < Elementary_File_Pointer > > 
      Elementary_File_Pointers ;

    typedef list < Data_Object_Pointer , 
                   persistent_allocator < Data_Object_Pointer > > 
      Data_Object_Pointers ;

    static const Memory::size_t initial_number_of_dedicated_file_slots  = 0 ;
    static const Memory::size_t initial_number_of_elementary_file_slots = 0 ;
    static const Memory::size_t initial_number_of_data_object_slots = 0 ;

    Dedicated_File ( const Dedicated_File_Pointer & parent_dedicated_file_pointer , 
                     const Octet * FCP_octets , 
                     Memory::size_t FCP_octets_size )
      : File ( parent_dedicated_file_pointer , FCP_octets , FCP_octets_size ) , 
        the_dedicated_file_pointers  ( initial_number_of_dedicated_file_slots  ) , 
        the_elementary_file_pointers ( initial_number_of_elementary_file_slots ) ,
        the_data_object_pointers ( initial_number_of_data_object_slots ) 
      { } ;

    Dedicated_File ( const Dedicated_File_Pointer & parent_dedicated_file_pointer , 
                     const Octet_String & FCP_octet_string )
      : File ( parent_dedicated_file_pointer , FCP_octet_string ) , 
        the_dedicated_file_pointers  ( initial_number_of_dedicated_file_slots  ) , 
        the_elementary_file_pointers ( initial_number_of_elementary_file_slots ) ,
        the_data_object_pointers ( initial_number_of_data_object_slots ) 
      { } ;


    File::Type find ( file_identifier_t file_identifier , 
                      Dedicated_File_Pointer & dedicated_file_pointer ,
                      Elementary_File_Pointer & elementary_file_pointer )
      {
        for ( Dedicated_File_Pointers::iterator pd = the_dedicated_file_pointers . begin ( ) ;
              pd != the_dedicated_file_pointers . end ( ) ;
              pd ++ )
          if ( file_identifier == ( * pd ) -> get_FCP ( ) . file_identifier ( ) )
            {
              dedicated_file_pointer = * pd ;
              elementary_file_pointer = Elementary_File_Pointer ( ) ;
              return File::dedicated ;
            }
        
        for ( Elementary_File_Pointers::iterator pd = the_elementary_file_pointers . begin ( ) ;
              pd != the_elementary_file_pointers . end ( ) ;
              pd ++ )
          if ( file_identifier == ( * pd ) -> get_FCP ( ) . file_identifier ( ) )
            {
              dedicated_file_pointer = 
                Dedicated_File_Pointer::object_pointer_to_indexed_pointer ( this ) ;
              elementary_file_pointer = * pd ;
              return File::elementary ;
            }

        return File::none ;
      } ;

    File::Type find ( file_identifier_t file_identifier )
      {
        Dedicated_File_Pointer dedicated_file_pointer ;
        Elementary_File_Pointer elementary_file_pointer ;
        return 
          find ( file_identifier , 
                 dedicated_file_pointer , 
                 elementary_file_pointer ) ;
      } ;

    File::Type find ( const BER_TLV_Tag & tag , 
                      Dedicated_File_Pointer & dedicated_file_pointer ,
                      Data_Object_Pointer & data_object_pointer )
      {
        Primitive_BER_TLV_Pointer the_tag_list ;

        for ( Data_Object_Pointers::iterator pd = the_data_object_pointers . begin ( ) ;
              pd != the_data_object_pointers . end ( ) ;
              pd ++ )
        {
            the_tag_list = ( * pd ) -> get_FCP ( ) . tag_list_pointer ( ) ;

            if ( memcmp ( the_tag_list -> values ( ) . data ( ) ,
                 tag . data ( ) , tag . length ( ) ) == 0 )
            {
                data_object_pointer = * pd ;
                return File::data_object ;
            }
        }

        return File::none ;
      } ;

    void detach_children ( )
      {
        for ( Dedicated_File_Pointers::iterator pd = the_dedicated_file_pointers . begin ( ) ;
              pd != the_dedicated_file_pointers . end ( ) ;
              pd ++ )
          {
             ( * pd ) -> the_parent_dedicated_file_pointer =
            Dedicated_File_Pointer ( ) ;
             ( * pd ) -> detach_children ( ) ;
          };
        
        for ( Elementary_File_Pointers::iterator pd = the_elementary_file_pointers . begin ( ) ;
              pd != the_elementary_file_pointers . end ( ) ;
              pd ++ )
          {
             ( * pd ) -> the_parent_dedicated_file_pointer =
               Dedicated_File_Pointer ( ) ;
          };
      }

    void dump ( File_System * the_file_system_pointer , unsigned int indentation ) ;

    Dedicated_File_Pointers  the_dedicated_file_pointers  ;
    Elementary_File_Pointers the_elementary_file_pointers ;
    Data_Object_Pointers     the_data_object_pointers ;
  } ;



class FILE_SYSTEM_DLL File_System : public Object
  {
  public:

    File_System ( const Octet_String & master_file_FCP_octet_string )
      : the_master_file_pointer 
          ( Dedicated_File_Pointer::object_pointer_to_indexed_pointer 
              ( new Dedicated_File ( Dedicated_File_Pointer ( ) , 
                                     master_file_FCP_octet_string ) ) )
      { reset ( ) ; } ;

    File_System ( const Octet * master_file_FCP_octets , size_t master_file_FCP_octets_size  )
      : the_master_file_pointer 
          ( Dedicated_File_Pointer::object_pointer_to_indexed_pointer 
              ( new Dedicated_File ( Dedicated_File_Pointer ( ) , 
                                     master_file_FCP_octets , 
                                     master_file_FCP_octets_size ) ) )
      { reset ( ) ; } ;

    // Break backlinks on deletion
    ~File_System ( )
      { the_master_file_pointer -> detach_children ( ) ; } ;

    void verify_current_elementary_file ( )
      {
        if ( the_current_elementary_file_pointer 
               == Elementary_File_Pointer ( ) )
          throw RAPDU_Packet::SW_NO_CURRENT_FILE ;
      } ;

    void verify_current_dedicated_file ( )
      {
        if ( the_current_elementary_file_pointer 
               != Elementary_File_Pointer ( ) )
          throw RAPDU_Packet::SW_INCOMPATIBLE_FILE_STRUCTURE ;
      } ;

    void select_file ( const file_identifier_t file_identifier , const Octet restriction )
      {
        switch ( file_identifier )
          {
            case MF:
              switch ( restriction )
                {
                  case 0x00:  // Explicit
                    break ;
                  case 0x01:  // Child DF
                    break ;
                  case 0x02:  // Child EF
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                  case 0x03:  // Parent DF
                    break ;
                  default:
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                }
              reset ( ) ;
              break ;

            case CURRENT_EF:
              switch ( restriction )
                {
                  case 0x00:  // Explicit
                    break ;
                  case 0x01:  // Child DF
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                  case 0x02:  // Child EF
                    break ;
                  case 0x03:  // Parent DF
                    break ;
                  default:
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                }
              verify_current_elementary_file ( ) ;
              break ;

            case PARENT_DIR:
              switch ( restriction )
                {
                  case 0x00:  // Explicit
                    break ;
                  case 0x01:  // Child DF
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                  case 0x02:  // Child EF
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                  case 0x03:  // Parent DF
                    break ;
                  default:
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                }
              if ( the_current_dedicated_file_pointer 
                     == the_master_file_pointer )
                throw RAPDU_Packet::SW_FILE_NOT_FOUND ;
              the_current_dedicated_file_pointer =
                the_current_dedicated_file_pointer 
                  -> the_parent_dedicated_file_pointer ;

              the_current_elementary_file_pointer =
                Elementary_File_Pointer ( ) ;
              break ;
 
            case CURRENT_DIR:
              switch ( restriction )
                {
                  case 0x00:  // Explicit
                    break ;
                  case 0x01:  // Child DF
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                  case 0x02:  // Child EF
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                  case 0x03:  // Parent DF
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                  default:
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                }

              the_current_elementary_file_pointer =
                Elementary_File_Pointer ( ) ;
              break ;

            case CURRENT_ADF:
              // ADF not implemented here -- up in File Manager?
              throw RAPDU_Packet::SW_FILE_NOT_FOUND ;
              break ;

            default:
              if ( File::none == 
                     the_current_dedicated_file_pointer 
                       -> find ( file_identifier , 
                                 the_current_dedicated_file_pointer , 
                                 the_current_elementary_file_pointer ) )
                throw RAPDU_Packet::SW_FILE_NOT_FOUND ;

              // Make sure we found the right kind of file
              switch ( restriction )
                {
                  case 0x00:  // Explicit
                    break ;

                  case 0x01:  // Child DF

                    if ( the_current_elementary_file_pointer 
                          != Elementary_File_Pointer ( ) )
                      throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                    break ;

                  case 0x02:  // Child EF

                    if ( the_current_elementary_file_pointer 
                          == Elementary_File_Pointer ( ) )
                      throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                    break ;
                  
                  case 0x03:  // Parent DF

                    if ( the_current_elementary_file_pointer 
                          != Elementary_File_Pointer ( ) )
                      throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                    break ;

                  default:
                    throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
                }
          }
      } ;

    void set_current_data_object_pointer ( const BER_TLV_Tag & tag )
    {
        Dedicated_File_Pointer dedicated_file_pointer ;

        if ( the_current_dedicated_file_pointer -> find ( tag , 
                                                     dedicated_file_pointer , 
                                                     the_current_data_object_pointer ) == File::none )
             throw RAPDU_Packet::SW_DATA_OBJECT_NOT_FOUND ;
      }

    File * get_current_file_pointer ( )
    {
        if ( the_current_elementary_file_pointer 
               == Elementary_File_Pointer ( ) )
            return dynamic_cast < File * > 
                     ( & * the_current_dedicated_file_pointer  ) ;
        else
            return dynamic_cast < File * > 
                     ( & * the_current_elementary_file_pointer ) ;
      }

    void create_file ( const Octet * FCP_octets , size_t FCP_octets_size )
      {
        create_file ( Octet_String ( FCP_octets , FCP_octets_size ) ) ;
      } ;

    void create_file ( const Octet_String & FCP_octet_string )
      {
        File_Control_Parameters the_FCP ( FCP_octet_string ) ;

        if ( File::none != the_current_dedicated_file_pointer 
                            -> find ( the_FCP . file_identifier ( ) ) )
                            throw RAPDU_Packet::SW_FILE_ALREADY_EXISTS ;

        if ( the_FCP . denotes_a_dedicated_file ( ) )
          {
            Dedicated_File_Pointer the_new_dedicated_file_pointer  =
              Dedicated_File_Pointer::object_pointer_to_indexed_pointer 
                ( new Dedicated_File ( the_current_dedicated_file_pointer , 
                                      FCP_octet_string ) ) ;

            the_current_dedicated_file_pointer 
              -> the_dedicated_file_pointers . push_front
                  ( the_new_dedicated_file_pointer ) ;

            the_current_dedicated_file_pointer  = 
              the_new_dedicated_file_pointer ;

            the_current_elementary_file_pointer = 
              Elementary_File_Pointer::null ( ) ;

            the_current_data_object_pointer = 
              Data_Object_Pointer::null ( ) ;
          }
        else
          {
            Memory::size_t initial_data_size = the_FCP . file_allocation ( );
            Elementary_File_Pointer the_new_elementary_file_pointer  =
              Elementary_File_Pointer::object_pointer_to_indexed_pointer 
                ( new Elementary_File ( the_current_dedicated_file_pointer , 
                                        FCP_octet_string , 
                                        initial_data_size ) ) ;

            the_current_dedicated_file_pointer 
              -> the_elementary_file_pointers . push_front 
                  ( the_new_elementary_file_pointer ) ;

            the_current_elementary_file_pointer = 
              the_new_elementary_file_pointer ;
          }
      } ;

    void create_data_object ( const Octet * FCP_octets , size_t FCP_octets_size )
      {
        create_data_object ( Octet_String ( FCP_octets , FCP_octets_size ) ) ;
      } ;

    void create_data_object ( const Octet_String & FCP_octet_string )
      {
        File_Control_Parameters the_FCP ( FCP_octet_string ) ;

        if ( File::none != the_current_dedicated_file_pointer 
                            -> find ( the_FCP . file_identifier ( ) ) )
                            throw RAPDU_Packet::SW_FILE_ALREADY_EXISTS ;
            
        Memory::size_t initial_data_size = the_FCP . file_allocation ( );
        Data_Object_Pointer the_new_data_object_pointer  =
            Data_Object_Pointer::object_pointer_to_indexed_pointer 
            ( new Data_Object ( the_current_dedicated_file_pointer , 
                                FCP_octet_string , 
                                initial_data_size ) ) ;

        the_current_dedicated_file_pointer 
            -> the_data_object_pointers . push_front 
                ( the_new_data_object_pointer ) ;

        the_current_data_object_pointer = 
            the_new_data_object_pointer ;
      } ;

    void update_binary ( const Octet_String & data , Memory::offset_t offset )
      {
        try
          {
            Octets::Pointer the_file_data = the_current_elementary_file_pointer -> the_data ;
            size_t the_file_data_size =  the_file_data -> get_size ( ) ;
            Memory::size_t length = data . length ( ) ;

            if ( the_file_data_size <= offset )
              throw RAPDU_Packet::SW_WRONG_P1_P2 ;

            if ( the_file_data_size < offset + length )
              throw RAPDU_Packet::SW_WRONG_LENGTH ;

            Octet_String::cpy ( the_file_data , 
                                offset , 
                                data ) ;
          }
        catch ( Octets::Index_Exception )
          {
            throw RAPDU_Packet::SW_WRONG_P1_P2 ;
          }
      } ;

    void read_binary ( Octet_String & data , Memory::offset_t offset , Memory::size_t length )
      {
        if ( 0 == length )
          throw RAPDU_Packet::SW_WRONG_LENGTH ;

        try
          {
            Octets::Pointer the_file_data = the_current_elementary_file_pointer -> the_data ;
            size_t the_file_data_size =  the_file_data -> get_size ( ) ;
            
            if ( the_file_data_size <= offset )
              throw RAPDU_Packet::SW_WRONG_P1_P2 ;

            if ( the_file_data_size < offset + length )
              throw RAPDU_Packet::SW_END_OF_FILE ;

            Octet_String::cpy ( data ,
                                the_file_data , 
                                offset , 
                                length ) ;
          }
        catch ( Octets::Index_Exception )
          {
            throw RAPDU_Packet::SW_WRONG_LENGTH ;
          }
      } ;

    void get_data ( const BER_TLV_Tag & tag ,
                    Octet_String & data_object )
      {
        set_current_data_object_pointer ( tag );

        try
          {
            Octets::Pointer the_data_object = the_current_data_object_pointer -> the_data ;
            size_t the_data_object_size =  the_data_object -> get_size ( ) ;

            data_object = Octet_String ( the_data_object , 
                                0 , 
                                the_data_object_size ) ;
          }
        catch ( Octets::Index_Exception )
          {
            throw RAPDU_Packet::SW_WRONG_LENGTH ;
          }
      } ;

    void put_data ( const BER_TLV_Tag & tag ,
                    const Octet_String & data )
      {
		unsigned int max_data_object_length = 0;

        set_current_data_object_pointer ( tag );
		
		switch( tag . get_octet ( 2 ) )
		{
			case 0x01:
				max_data_object_length = DO_PIV_X509_SIZE;
				break;
			case 0x02:
				max_data_object_length = DO_CHUID_SIZE;
				break;
			case 0x03:
				max_data_object_length = DO_FINGER_1_SIZE;
				break;
			case 0x04:
				max_data_object_length = DO_FINGER_2_SIZE;
				break;
			case 0x05:
				max_data_object_length = DO_CARD_X509_SIZE;
				break;
			case 0x06:
				max_data_object_length = DO_SECURITY_OBJECT_SIZE;
				break;
			case 0x07:
				max_data_object_length = DO_CCC_SIZE;
				break;
			case 0x08:
				max_data_object_length = DO_FACIAL_IMAGE_SIZE;
				break;
			case 0x09:
				max_data_object_length = DO_PRINTED_INFO_SIZE;
				break;
			case 0x0A:
				max_data_object_length = DO_SIGNING_X509_SIZE;
				break;
			case 0x0B:                
				max_data_object_length = DO_KEY_MANAGEMENT_X509_SIZE;
				break;
			default:
				max_data_object_length = 0;
		}

		if ( data . length ( ) > max_data_object_length )
			throw RAPDU_Packet::SW_NOT_ENOUGH_SPACE ;

        try
          {
            Octets::Pointer the_old_data = the_current_data_object_pointer -> the_data ;

            Octets::Pointer the_new_data ( Octets::alloc ( data . length ( ) ) ) ;
            memcpy(the_new_data -> body( ) , data . data( ) , data . length ( ) ) ; 

            the_current_data_object_pointer -> the_data = the_new_data ;

            Octets::free( the_old_data );
          }
        catch ( Octets::Index_Exception )
          {
            throw RAPDU_Packet::SW_WRONG_P1_P2 ;
          }
      } ;

    void reset ( )
      {
        the_current_dedicated_file_pointer = the_master_file_pointer ;
        the_current_elementary_file_pointer = Elementary_File_Pointer::null ( ) ;
        the_current_data_object_pointer = Data_Object_Pointer::null ( ) ;
      } ;

    void dump ( ) ;

    bool is_current_file ( const Dedicated_File & dedicated_file )
      {
        return the_current_elementary_file_pointer == Elementary_File_Pointer::null ( ) 
                 && ( & dedicated_file ) == ( & * the_current_dedicated_file_pointer ) ;
      } ;

    bool is_current_file ( const Elementary_File & elementary_file )
      {
        return the_current_elementary_file_pointer != Elementary_File_Pointer::null ( ) 
                 && ( & elementary_file ) == ( & * the_current_elementary_file_pointer ) ;
      } ;

    void verify_authorization ( const Security_Attribute_Pointer & security_attribute_pointer , 
                                Octet access_mode ) ;
    void verify_elementary_file_authorization ( Octet access_mode ) ;
    void verify_dedicated_file_authorization ( Octet access_mode ) ;
    void verify_data_object_authorization ( const BER_TLV_Tag & object_identifier,
                                            Octet access_mode );
    void verify_data_object_authorization ( Octet access_mode ) ;

    Dedicated_File_Pointer  the_current_dedicated_file_pointer ;
    Elementary_File_Pointer the_current_elementary_file_pointer ;
    Data_Object_Pointer the_current_data_object_pointer ;
    Dedicated_File_Pointer  the_master_file_pointer ;

    Security_System_Pointer the_security_system_pointer ;
  } ;

#endif  // !defined(__FILE_SYSTEM_H_INCLUDED__)