// SocketIO -- Socket Setup and Teardown support
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include <Messages.h>
#include "SocketIO.h"

#if defined(_WIN32)

#include <Ws2tcpip.h>

#define WINSOCK_2_0_NOT_SUPPORTED      ( WSABASEERR - 1)

#else

#error TODO: Non-Windows socket include(s)?

#endif

static int SocketIO_initialized = 0 ;


static void PrintSocketIOError(const long error_code)
{
  const char * description = "unknown error" ;
#if defined(_WIN32)
  //  Windows Sockets Error Codes
  const struct { long error_code ; const char * description ; } error_descriptions [ ] = 
    {
      { WSAEINTR              , "Interrupted function call."                              } ,
      { WSAEACCES             , "Permission denied."                                      } ,
      { WSAEFAULT             , "Bad address."                                            } ,
      { WSAEINVAL             , "Invalid argument."                                       } ,
      { WSAEMFILE             , "Too many open files."                                    } ,
      { WSAEWOULDBLOCK        , "Resource temporarily unavailable."                       } ,
      { WSAEINPROGRESS        , "Operation now in progress."                              } ,
      { WSAEALREADY           , "Operation already in progress."                          } ,
      { WSAENOTSOCK           , "Socket operation on nonsocket."                          } ,
      { WSAEDESTADDRREQ       , "Destination address required."                           } ,
      { WSAEMSGSIZE           , "Message too long. "                                      } ,
      { WSAEPROTOTYPE         , "Protocol wrong type for socket."                         } ,
      { WSAENOPROTOOPT        , "Bad protocol option."                                    } ,
      { WSAEPROTONOSUPPORT    , "Protocol not supported."                                 } ,
      { WSAESOCKTNOSUPPORT    , "Socket type not supported."                              } ,
      { WSAEOPNOTSUPP         , "Operation not supported."                                } ,
      { WSAEPFNOSUPPORT       , "Protocol family not supported."                          } ,
      { WSAEAFNOSUPPORT       , "Address family not supported by protocol family."        } ,
      { WSAEADDRINUSE         , "Address already in use."                                 } ,
      { WSAEADDRNOTAVAIL      , "Cannot assign requested address."                        } ,
      { WSAENETDOWN           , "Network is down."                                        } ,
      { WSAENETUNREACH        , "Network is unreachable."                                 } ,
      { WSAENETRESET          , "Network dropped connection on reset."                    } ,
      { WSAECONNABORTED       , "Software caused connection abort."                       } ,
      { WSAECONNRESET         , "Connection reset by peer."                               } ,
      { WSAENOBUFS            , "No buffer space available."                              } ,
      { WSAEISCONN            , "Socket is already connected."                            } ,
      { WSAENOTCONN           , "Socket is not connected."                                } ,
      { WSAESHUTDOWN          , "Cannot send after socket shutdown."                      } ,
      { WSAETIMEDOUT          , "Connection timed out."                                   } ,
      { WSAECONNREFUSED       , "Connection refused."                                     } ,
      { WSAEHOSTDOWN          , "Host is down."                                           } ,
      { WSAEHOSTUNREACH       , "No route to host."                                       } ,
      { WSAEPROCLIM           , "Too many processes."                                     } ,
      { WSASYSNOTREADY        , "Network subsystem is unavailable."                       } ,
      { WSAVERNOTSUPPORTED    , "Winsock.dll version out of range."                       } ,
      { WSANOTINITIALISED     , "Successful WSAStartup not yet performed."                } ,
      { WSAEDISCON            , "Graceful shutdown in progress."                          } ,
      { WSATYPE_NOT_FOUND     , "Class type not found."                                   } ,
      { WSAHOST_NOT_FOUND     , "Host not found."                                         } ,
      { WSATRY_AGAIN          , "Nonauthoritative host not found."                        } ,
      { WSANO_RECOVERY        , "This is a nonrecoverable error."                         } ,
      { WSANO_DATA            , "Valid name, no data record of requested type."           } ,
      { WSA_INVALID_HANDLE    , "Specified event object handle is invalid."               } ,
      { WSA_INVALID_PARAMETER , "One or more parameters are invalid."                     } ,
      { WSA_IO_INCOMPLETE     , "Overlapped I/O event object not in signaled state."      } ,
      { WSA_IO_PENDING        , "Overlapped operations will complete later."              } ,
      { WSA_NOT_ENOUGH_MEMORY , "Insufficient memory available."                          } ,
      { WSA_OPERATION_ABORTED , "Overlapped operation aborted."                           } ,
      //  { WSAINVALIDPROCTABLE   , "Invalid procedure table from service provider."          } ,
      //  { WSAINVALIDPROVIDER    , "Invalid service provider version number."                } ,
      //  { WSAPROVIDERFAILEDINIT , "Unable to initialize a service provider."                } ,
      { WSASYSCALLFAILURE     , "System call failure."                                    } 
    } ;

  int i ;

  for ( i = 0 ; 
        i < sizeof ( error_descriptions ) / sizeof ( error_descriptions [ 0 ] ) ; 
        i ++ )
    if ( error_code == error_descriptions [ i ] . error_code )
      {
        description = error_descriptions [ i ] . description ;
        break ;
      }

#else

#error TODO: Non-Windows socket initialization?

#endif

  error_message ( "Socket error %ld: %s", error_code , description );
}


long SocketIO_initialize ( )
{
  long Result = 0 ;
 
  if ( ! SocketIO_initialized )
    {
#if defined(_WIN32)

      WSADATA WinSockAPIData;
      const unsigned short int WinSockAPIVersionRequested = MAKEWORD( 2 , 0 ); // We need 2.0 functionality 
 
      Result = WSAStartup( WinSockAPIVersionRequested, &WinSockAPIData );
      if ( NO_ERROR != Result  ) 
        {
          PrintSocketIOError(Result);
          goto cleanup ;
        }
      // Confirm that the WinSock DLL supports 2.0.
      // Note that if the DLL supports versions greater    
      // than 2.0 in addition to 2.0, it will still return 
      // 2.0 in wVersion since that is the version we      
      // requested.                                        
      
      if ( WinSockAPIData.wVersion != WinSockAPIVersionRequested )
        {
          Result = WINSOCK_2_0_NOT_SUPPORTED ;
          PrintSocketIOError(Result);
          goto cleanup ;
        }
 
      // The WinSock DLL is acceptable.   
#else

#error TODO: Non-Windows socket initialization?

#endif
      SocketIO_initialized = 1 ;
    }
    return ( Result ) ;

cleanup:
    SocketIO_finalize ( ) ;
    return ( Result ) ;

}



long SocketIO_initialize_listener ( int port , SOCKET * listening_socket , SOCKET * connection_socket )
{
  long Result = NO_ERROR ;
 
  struct sockaddr_in
    {
      u_short sin_family;
      u_short sin_port;
      u_long  sin_addr;
      char    sin_zero[8];
    } name =
    {
      AF_INET ,
      htons(port) ,
      htonl(INADDR_ANY) ,
      0, 0, 0, 0, 0, 0, 0, 0
    } ;

  struct sockaddr_in cli_addr;
  int clilen = sizeof(cli_addr);


  if ( INVALID_SOCKET == ( * listening_socket = socket ( AF_INET , SOCK_STREAM , IPPROTO_IP ) ) ) 
    {
      Result = WSAGetLastError() ;
      PrintSocketIOError(Result);
      goto cleanup;
    }


	if ( 0 != bind ( * listening_socket , ( LPSOCKADDR ) &name, sizeof ( name ) ) )
    {
      Result = WSAGetLastError() ;
      PrintSocketIOError(Result);
      goto cleanup;
    }


  if ( SOCKET_ERROR == listen ( * listening_socket , 5 ) )
    {
      Result = WSAGetLastError() ;
      PrintSocketIOError(Result);
      goto cleanup;
    }

  if ( INVALID_SOCKET == 
       ( * connection_socket = 
           accept ( * listening_socket , (struct sockaddr *) &cli_addr, &clilen) ) )
    {
      Result = WSAGetLastError() ;
      PrintSocketIOError(Result);
      goto cleanup;
    }

    return ( Result ) ;

cleanup:
    SocketIO_finalize_listener ( listening_socket , connection_socket ) ;
    return ( Result ) ;

}

long SocketIO_initialize_talker ( const char * host , int port , SOCKET * talking_socket )
{ 
   long Result = NO_ERROR ;

   struct addrinfo * AddressInformationPtr ;

   typedef 
  struct sockaddr_in
    {
      u_short sin_family;
      u_short sin_port;
      u_long  sin_addr;
      char    sin_zero[8];
    } sockaddr_in ;


 sockaddr_in hostname =
    {
      AF_INET ,
      htons(port) ,
      htonl(INADDR_LOOPBACK) ,
      0, 0, 0, 0, 0, 0, 0, 0
    } ;


  sockaddr_in localname =
    {
      AF_INET ,
      0 ,
      0 ,
      0, 0, 0, 0, 0, 0, 0, 0
    } ;


  if ( NULL != host )
  {
     Result = getaddrinfo (host , NULL , NULL, & AddressInformationPtr ) ;
     if ( NO_ERROR != Result )
     {
         PrintSocketIOError(Result);
         goto cleanup;
     }
     hostname . sin_addr = 
        ((struct sockaddr_in *)(AddressInformationPtr ->ai_addr)) -> sin_addr ;
     freeaddrinfo(AddressInformationPtr);
  }


  if ( INVALID_SOCKET == ( * talking_socket = socket ( AF_INET , SOCK_STREAM , IPPROTO_IP ) ) ) 
    {
      Result = WSAGetLastError() ;
      PrintSocketIOError(Result);
      goto cleanup;
    }


	if ( 0 != bind ( * talking_socket , ( LPSOCKADDR ) &localname, sizeof ( localname ) ) )
    {
      Result = WSAGetLastError() ;
      PrintSocketIOError(Result);
      goto cleanup;
    }

   // FIXME: Use non-blocking connect and select with some (infinite?) timeout
  if ( SOCKET_ERROR == connect ( * talking_socket , ( LPSOCKADDR ) & hostname, sizeof ( hostname ) ) )
    {
      Result = WSAGetLastError() ;
      PrintSocketIOError(Result);
      goto cleanup;
    }


    return ( Result ) ;

cleanup:
    SocketIO_finalize_talker ( talking_socket ) ;
    return ( Result ) ;

}


long SocketIO_finalize ( )
{
  unsigned long int Result = 0 ;

  if ( SocketIO_initialized )
    {
#if defined(_WIN32)

      if ( SOCKET_ERROR == WSACleanup ( ) )
        {
            Result = WSAGetLastError() ;
            PrintSocketIOError(Result);
        }

#else

#error TODO: Non-Windows socket initialization?

#endif

       SocketIO_initialized = 0 ;
     }
  return ( Result ) ;
}


long SocketIO_finalize_listener ( SOCKET * listening_socket , SOCKET * connection_socket )
{
   unsigned long int Result = NO_ERROR ;

   if ( INVALID_SOCKET != * connection_socket )
   {
      if ( SOCKET_ERROR == shutdown ( * listening_socket , SD_SEND ) )
      {
         Result = WSAGetLastError() ;
         if ( WSAENOTCONN != Result )
           PrintSocketIOError(Result);
      }
      * connection_socket = INVALID_SOCKET ;
   }

   if ( INVALID_SOCKET != * listening_socket )
   {
      if ( SOCKET_ERROR == closesocket ( * listening_socket ) )
      {
         Result = WSAGetLastError() ;
         PrintSocketIOError(Result);
      }
      * listening_socket = INVALID_SOCKET ;
   }

   return ( Result ) ;
}


long SocketIO_finalize_talker ( SOCKET * talking_socket )
{
   unsigned long int Result = NO_ERROR ;

   if ( INVALID_SOCKET != * talking_socket )
   {
      if ( SOCKET_ERROR == closesocket ( * talking_socket ) )
      {
         Result = WSAGetLastError() ;
         PrintSocketIOError(Result);
      }
      * talking_socket = INVALID_SOCKET ;
   }

   return ( Result ) ;
}
