// TLP_APDU_Client -- TLP224 Listener
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include <TLP_APDU_Server.h>
#include <Octet_String.h>

#include <macros.h>

const char TLP_APDU_Server::default_hostname [ ]= "localhost" ;


long TLP_APDU_Server::initialize ( int v )
{
  int return_value = 1;
  
  debug_message ( "TLP_APDU_Server::initialize()\n" );
  
  // check for negative ints and set verbosity to 0 as a minimum value
  if ( v < 0 ) 
    v = 0;

  set_verbosity ( v );
  if ( NULL == hostname ) 
  {
    hostname = default_hostname ;
    port = DEFAULT_APDU_PORT;
  }
  
  // handle negative and 0 value for port, set to default
  if ( port <= 0 )
    port = DEFAULT_APDU_PORT;
 
  return ! ( the_TLP224 . initialize_socket_io ( card_edge_socket )
             && the_TLP224 . initialize_socket_client ( card_edge_socket, 
                                                             (const char *) hostname, 
                                                             (u_short) port ,
                                                             report_initialization_errors ,
                                                             true ) ) ;
}


long TLP_APDU_Server::exchange_APDUs ( const CAPDU_Packet & command, RAPDU_Packet & response )
{
  debug_message ( "TLP_APDU_Server::exchange_APDUs()\n" );
     
  size_t APDU_length = command.get_APDU_length();
  unsigned char rAPDU  [1024] ;            // TODO: make this dynamic too?
  int number_of_bytes_received;
  Octet value;

  debug_message ( "copying command:" );
  
  // Copy byte from command APDU to TLP-224 output buffer
  for ( unsigned int i = 0 ; i < APDU_length; i++ )
  {
    value = command.get_APDU_octet ( i );
    
    if ( 1 < verbosity ) 
      debug_message (" %02X", value );

    the_TLP224 . put_byte_into_output_buffer ( value );
  }

  debug_message ( "\n" 
                  "APDU_length=%d\n" , static_cast < unsigned int > ( APDU_length ) ) ;

  const Octet P3 = command .get_APDU_octet(CAPDU_Packet::OFFSET_P3) ;
  
  debug_message ( "P3 = 0x%02X\n" , ( unsigned int ) P3 );
  
	if ( APDU_length - 5 < command.get_APDU_octet ( 4 ) ) 
  {
    debug_message ( "Case 2 or Case 4, P3 interpreted as Le. (P3 < payload data length ). \n" );
    the_TLP224 . send_TLP224_packet    (  card_edge_socket, TLP224::TLPC_INPUT_FROM_MICROPROCESSOR , static_cast < int > ( APDU_length )  );
	} 
  else 
  {
    debug_message ( "Case 1 or Case 3, P3 interpreted as Lc. ( payload data length <= P3 ). \n" );
    the_TLP224 . send_TLP224_packet    (  card_edge_socket, TLP224::TLPC_OUTPUT_TO_MICROPROCESSOR , static_cast < int > ( APDU_length )  );
  }

  unsigned char tlp_command;
  if ( ! the_TLP224 . receive_TLP224_packet (  card_edge_socket, tlp_command , number_of_bytes_received ) )
    return 1 ;

  // fill up rAPDU buffer
  for ( int i=0; i < number_of_bytes_received; i++ )
    the_TLP224 . get_byte_from_input_buffer ( rAPDU [ i ] ) ;
  
  // handle the inline get response

  if ( 2 == number_of_bytes_received && 0x61 == rAPDU [ 0 ] )
  {
 	  debug_message ( "Received a 61XX respone, so we are going to get the rest of the bytes\n" );
			
	  Octet number_of_bytes_to_request = rAPDU[1];

		debug_message ( "number_of_bytes_to_request=%d\n" , number_of_bytes_to_request );

			
		const Octet a_get_response [] = 
      { 
        CLA_INS ( CAPDU_Packet::CLAINS_GET_RESPONSE) , 0x00, 0x00, number_of_bytes_to_request   // Cute, eh?
      };
      
    for ( int i = 0 ; i < sizeof ( a_get_response ); i++ )
      the_TLP224 . put_byte_into_output_buffer ( a_get_response [ i ] );
  
    // TLPC_INPUT_FROM_MICROPROCESSOR= 0xDB; 
    if ( ! the_TLP224 . send_TLP224_packet    (  card_edge_socket, 0xDB , sizeof ( a_get_response ) ) )
      return 1 ;

    if ( ! the_TLP224 . receive_TLP224_packet (  card_edge_socket, tlp_command , number_of_bytes_received ) )
      return 1 ;

    for ( int i=0; i < number_of_bytes_received; i++ )
      the_TLP224 . get_byte_from_input_buffer ( rAPDU [ i ] ) ;

  }

  response . set_APDU_length ( number_of_bytes_received ) ;
  for ( int i = 0 ; i < number_of_bytes_received ; i ++ )
    response . put_APDU_octet ( i , rAPDU [ i ] )  ;

  debug_message ( "received : %d bytes \n"
                  "returning 0\n" ,
                  number_of_bytes_received );

  return 0;
}



long TLP_APDU_Server::finalize ( )
{
  debug_message ( "TLP_APDU_Server::finalize()\n" ) ;

  the_TLP224 . finalize_socket_io ( card_edge_socket );
  
  return 0;
}



long TLP_APDU_Server::open_log_file    ( const char * filename ) 
  {
    debug_message ( "TLP_APDU_Server::open_log_file()\n" );
    return 1 ;
  }



 long TLP_APDU_Server::flush_log_file   ( void ) 
 {
   debug_message ( "TLP_APDU_Server::flush_log_file()\n" );

   return 1 ;
 }



long TLP_APDU_Server::close_log_file   ( void ) 
{
  debug_message ( "TLP_APDU_Server::close_log_file()\n" );

  return 1 ;
}



long TLP_APDU_Server::log_file_comment ( const char * comment ) 
{
  debug_message ( "TLP_APDU_Server::log_file_comment( %s )\n" , comment ) ;

  return 1 ;
}


void TLP_APDU_Server::set_verbosity    ( int v ) 
{
  debug_message ( "TLP_APDU_Server::set_verbosity (%d) \n" , v );
  verbosity = v;
}


int  TLP_APDU_Server::get_verbosity    ( void ) 
{
  debug_message ( "TLP_APDU_Server::get_verbosity()\n");
  return verbosity;
}



long TLP_APDU_Server::power_up         ( void )
{
  debug_message ( "TLP_APDU_Server::power_up()\n" ) ;

  unsigned char response [1024] ;
  int num_received;

  static const unsigned char SHORT_TLP_RESET_COMMAND [ ] = 
    { /*TLP_ACK*/ /*TLP_LEN=0x04*/ /*OSB= 0x6E*/ 0x00, 0x00, 0x00 /* BCC=0x0A */ /* ETX */ };

  for ( int i = 0 ; i < sizeof ( SHORT_TLP_RESET_COMMAND ); i++ )
    the_TLP224 . put_byte_into_output_buffer ( SHORT_TLP_RESET_COMMAND [ i ] );

  if ( 0 < verbosity )
    debug_message ( "TLP_APDU_Server::power_up: trying send_TLP224_packet(%04X) ...\n",
                    card_edge_socket ) ;

  if ( ! the_TLP224 . send_TLP224_packet    (  card_edge_socket, 0x6E , sizeof ( SHORT_TLP_RESET_COMMAND ) ) )
    {
      error_message ( "Connection broken during power up!\n" ) ;
      return 1 ;
    }

  if ( 0 < verbosity )
    debug_message ( "TLP_APDU_Server::power_up: send_TLP224_packet(%04X) succeeded \n"
                    "                           trying receive_TLP224_packet (%04X) ...\n",
                    card_edge_socket ,
                    card_edge_socket ) ;

  unsigned char tlp_command;
  if ( ! the_TLP224 . receive_TLP224_packet (  card_edge_socket, tlp_command , num_received ) )
    {
      error_message ( "Connection broken during power up!\n" ) ;
      return 1 ;
    }

  if ( 0 < verbosity )
    debug_message ( "TLP_APDU_Server::power_up: receive_TLP224_packet(%04X) succeeded\n",
                    card_edge_socket ) ;

  for ( int i=0; i < num_received; i++ )
    the_TLP224 . get_byte_from_input_buffer ( response [ i ] ) ;

  if ( ! ( 0x00 == tlp_command && 0x28 == response [ 0 ]  && 0x02 == response [ 1 ] 
           && num_received - 3 == response [ 2 ] ) )
    {
      error_message ( "Unexpected response from power up!\n" ) ;
      return 1 ;
    }

  the_ATR_length = num_received - 3 ;
  memcpy ( the_ATR , & response [ 3 ] , the_ATR_length ) ;

  if ( 1 < verbosity ) 
  {
    debug_message ( " received : %d bytes \n" , num_received );
    debug_message ( "returning 0\n" );
  }

  return 0 ;
}


long  TLP_APDU_Server::power_down       ( void ) 
{
  debug_message ( "TLP_APDU_Server::power_down()\n" );
  
  the_TLP224 . send_TLP224_packet    (  card_edge_socket, 0x4D, 0x00 );  // 0x4D == TLP POWER DOWN COMMAND
  // No response packet!

  debug_message ( "returning 0\n" );

  return 0 ;
}



int TLP_APDU_Server::set_hostname ( const char * passed_hostname )
{
  debug_message ( "TLP_APDU_Server::set_hostname( \"%s\" )\n" , passed_hostname ) ; 

   // This seems *WRONG*, I really want to be passing references to String objects
 
   // The real code should do this:
   // 1) Check the length of the input string, if it is unreasonably long then die or truncate
   // 2) Check the length of the destination buffer, if it is too short, re-create it to length (if length is reasonable )
   // 3) perform step 1 again

  if ( strlen ( passed_hostname ) < MAX_HOSTNAME_BUFFER_LENGTH )  // check if the length is reasonable
  {
    // recreate buffer
    hostname = strdup ( passed_hostname ) ; 
    return 0 ;
  }

  return 1 ;  // Didn't know how to handle it.
}


const char * TLP_APDU_Server::get_hostname ( void )
{
  debug_message ( "TLP_APDU_Server::get_hostname() -- returning:\"%s\"\n" , hostname );

  return hostname;
}


void TLP_APDU_Server::get_ATR( unsigned char * ATR , 
                               unsigned int & ATR_length )
  {
    ATR_length = the_ATR_length ;
    memcpy ( ATR , the_ATR , the_ATR_length ) ; 
  } ;


bool TLP_APDU_Server::store_persistent_state ( FILE * , unsigned short & )
  { return true ; } ; 

bool TLP_APDU_Server::load_persistent_state ( FILE * , unsigned short & )
  { return true ; } ; 
