// TLP224 Input and Output
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include <TLP224.h>
#include <hexify.h>
#include <Messages.h>

#if defined(_WIN32)
#include <ws2tcpip.h>
#else
#error Non-windows include(s)?
#endif

bool TLP224::debug_TLP224_traffic = false ;  // Set to true to see all the TLP-224 traffic

TLP224::TLP224()
{
  buffer_end = buffer ;
  buffer_pointer = buffer_end ;
  power_on = false ;
  output_buffer_end = output_buffer + sizeof(output_buffer);
  output_buffer_pointer = output_buffer ;
  input_buffer_end = input_buffer ;
  input_buffer_pointer = input_buffer_end ;
}


#ifdef DOCUMENTATION
status_code_description StatusCodes [] = 
{
   //Code   Meaning
   { 0x00 , "OK." } ,
   { 0x01 , "Unknown driver or command." } ,
   { 0x02 , "Operation not possible with this driver." } ,
   { 0x03 , "Incorrect number of arguments." } ,
   { 0x04 , "Command unknown. The first byte of the command is not a valid command code." } ,
   { 0x05 , "Response too long for the buffer." } ,
   { 0x09 , "Communication protocol error. The header of a message is neither ACK or NACK (60h or E0h)" } ,
   { 0x10 , "Response error at the card reset. The first byte of the response (TS) is not valid 11h ISO command header error. The byte INS in the ISO header is not valid (6x or 9x)." } ,
   { 0x12 , "Message too long. The buffer is limited to 254 bytes, of which 248 bytes are for the data exchanged with the card" } ,
   { 0x13 , "Byte reading error returned by an asynchronous card. " } ,
   { 0x15 , "Card turned off. A Power Up command must be applied to the card prior to any other operation." } ,
   { 0x16 , "Programming voltage not available. The parameter V in the DEFINE CARD TYPE command is not valid." } ,
   { 0x17 , "Communication protocol unknown or incorrectly initialized." } ,
   { 0x18 , "Communication protocol unknown or incorrectly initialized." } ,
   { 0x19 , "Illegal access to external bus." } ,
   { 0x1A , "Error in an ISO format card command. The parameter LN in the ISO header does not correspond to the actual length of the data." } ,
   { 0x1B , "A command has been sent with an incorrect number of parameters." } ,
   { 0x1D , "The check byte TCK of the response to reset of a microprocessor card is incorrect." } ,
   { 0x1E , "An attempt has been made to write to external memory, which is write protected." } ,
   { 0x1F , "Incorrect data has been sent to the external memory. This error is returned after a write check during a downloading operation." } ,
   { 0xA0 , "Error in the card reset response, such as unknown exchange protocol, or byte TA1 not recognized. The card is not supported. The card reset response is nevertheless returned." } ,
   { 0xA1 , "Card protocol error (T=0/T=1)." } ,
   { 0xA2 , "Card malfunction. The card does not respond to the reset or has interrupted an exchange (by time-out)." } ,
   { 0xA3 , "Parity error (in the course of an exchange microprocessor). The error only occurs after several unsuccessful attempts at re transmission." } ,
   { 0xA4 , "Card has aborted chaining (T=1)." } ,
   { 0xA5 , "Chaining aborted. (T=1)." } ,
   { 0xA6 , "Protocol type Selection (PTS) error." } ,
   { 0xCF , "Overkey already pressed." } ,
   { 0xE4 , "The card has just sent an invalid \"Procedure Byte\" (see ISO 7816-3)." } ,
   { 0xE5 , "The card has interrupted an exchange (the card sends an SW1 byte but more data has to be sent or received)." } ,
   { 0xE7 , "Error returned by the card. The bytes SW1 and SW2 returned by the card are different from 90h 00." } ,
   { 0xF7 , "Card removed. The card has been withdrawn in the course of carrying out of a command. Check that the card instruction is not partially completed." } ,
   { 0xF8 , "The card is consuming too much electricity or is short circuiting." } ,
   { 0xFB , "Card absent. There is no card in the smart card interface. The card may have been removed when it was powered up, but no command has been interrupted." } ,
};
#endif // DOCUMENTATION


void TLP224::analyze_socket_io_error ( )
{
  char error_msg_buffer[257] = "Unknown error ID!";
#if defined(_WIN32)
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, WSAGetLastError ( ) , 0 , error_msg_buffer, 256, NULL );
#else
#error Non-windows error message formatting?
#endif
  error_message (  "%s\n" , error_msg_buffer) ;
}



bool TLP224::recv_buffer ( SOCKET & connection_socket )
{
  if ( INVALID_SOCKET == connection_socket )
    return false ;

  // TODO: The line below is a surrogate for the proper code, which
  // I think ought to recv until it does get a zero return, thereby
  // picking up the whole packet.
  // I.e., the algorithm should be
  //   sleep a bit
  //   if recv() returns zero, we're done
  //   else
  //     do
  //       sleep a bit more
  //       try recv again, appending to the buffer
  //     until the recv returns zero

  Sleep ( 100 );  // Need this to give CREF time to send a decent response.

  int recv_result = recv ( connection_socket , (char *)buffer , 4096 , 0 ) ;

  if ( 0 == recv_result ) /* connection closed */
    {
      error_message (  "Socket connection closed.\n" );
      return false ;
    }

  if ( SOCKET_ERROR == recv_result )
    {
      analyze_socket_io_error ( ) ;
      connection_socket = INVALID_SOCKET ;
      return false;
    }

  if ( 0 == recv_result % 2 )
    {
      error_message (  "Packet does not contain odd number of characters in recv_buffer!\n" ) ;
      return false ;
    }

  buffer_end = buffer + recv_result - 1 ;

  for ( buffer_pointer = buffer ;
        buffer_pointer < buffer_end ;
        buffer_pointer ++ )
    if ( ! ishex ( * buffer_pointer ) )
      {
        error_message (  "Illegal character '\\x%02X' in packet in recv_buffer!\n", * buffer_pointer ) ;
        return false ;
      }

  if ( '\x03' !=  * buffer_pointer )
    {
      error_message (  "Packet does not end with ETX in recv_buffer!\n" ) ;
      return false ;
    }

  if ( debug_TLP224_traffic ) 
    {
      * buffer_pointer = '\x00' ;
      debug_message ( "recv_buffer length=%d buffer:%s\\x03\n" , recv_result , buffer ) ;
    } 
  * buffer_pointer = '\x03' ;

  buffer_pointer = buffer ;
  
  return true ;
}


bool TLP224::send_buffer ( SOCKET & connection_socket )
{
  if ( INVALID_SOCKET == connection_socket )
    return false ;

  if ( 0 !=  ( buffer_pointer - buffer ) % 2 )
    {
      error_message (  "Packet does not contain even number of characters in send_buffer!\n" ) ;
      return false ;
    }

  if ( debug_TLP224_traffic ) 
    {
      * buffer_pointer = '\x00' ;
      debug_message ( "send_buffer length=%d buffer:%s\\x03\n" , buffer_pointer - buffer - 1 , buffer ) ;
    } 
  * buffer_pointer ++ = '\x03' ;

  _flushall();

  int send_result = send ( connection_socket , (const char *)buffer , (int)( buffer_pointer - buffer ) , 0 ) ;

  if ( 0 == send_result ) /* connection closed */
    {
      error_message (  "Socket connection unexpectedly closed.\n");
      return false ;
    }

  if ( SOCKET_ERROR == send_result )
    {
      analyze_socket_io_error ( ) ;
      connection_socket = INVALID_SOCKET ;
      return false;
    }

  buffer_pointer = buffer_end = buffer ;
  
  return true ;
}


bool TLP224::get_socket_byte ( unsigned char & b )
{
  if ( buffer_end == buffer_pointer )
    return false ;

  char c1 , c2 ;

  c1 = * buffer_pointer ++ ;
  c2 = * buffer_pointer ++ ;

  b = hexvalue ( c1 ) << 4 | hexvalue ( c2 ) ;
  return true ;
}


void TLP224::put_socket_byte ( unsigned char b )
{
  * buffer_pointer ++ = hexchar( b >> 4 ) ;
  * buffer_pointer ++ = hexchar( b & 15 ) ;
}


bool TLP224::receive_TLP224_packet ( SOCKET & connection_socket , unsigned char & command , int & length )
{
  unsigned char b , BCC ;
  int i ;

  if ( ! recv_buffer ( connection_socket ) )
    return false ;

  if ( ! get_socket_byte ( b ) )
    return false ;
  else
    BCC = b ;
  if ( b != ((unsigned char)TLP_ACK) )
    {
      error_message (  "Unexpected TYPE value of %02X\n" , b ) ;
      return false ;
    }

  if ( ! get_socket_byte ( b ) )
    return false ;
  else
    BCC ^= b ;
  length = b ;

  if ( 0 == length )
    return false ;

  length -- ;

  if ( ! get_socket_byte ( command ) )
    return false ;
  else
    BCC ^= command ;
  
  input_buffer_end = input_buffer ;

  for ( i = 0 ; i < length ; i ++ ) 
    {
      if ( ! get_socket_byte ( b ) )
        return false ;
      else
        BCC ^= b;
        * input_buffer_end ++ = b;
    }

  if ( ! get_socket_byte ( b ) )
    return false ;

  if ( BCC != b )
    {
      error_message (  "Checksum error\n" ) ;
      return false ;
    }

  if ( get_socket_byte ( b ) )
    {
      error_message (  "Packet length error\n" ) ;
      return false ;
    }

  if ( debug_TLP224_traffic ) 
    {
      debug_message ( "receive_TLP224_packet length %d command %02X:" , length , command ) ;
      for ( i = 0 ; i < length ; i ++ ) 
        debug_message ( " %02X" , (unsigned char) input_buffer [ i ] ) ;
      debug_message ( "\n" ) ;
    } 

  input_buffer_pointer = input_buffer ;

  return true ;
}





bool TLP224::send_TLP224_packet ( SOCKET & connection_socket , unsigned char OSB , int length )
{
  int i ;

  if ( debug_TLP224_traffic ) 
    {
      debug_message ( "send_TLP224_packet OSB %02X length %d output_buffer:" , OSB , length ) ;
      for ( i = 0 ; i < length ; i ++ ) 
        {
          debug_message ( " %02X" , (unsigned char) output_buffer [ i ] ) ;
        }
      debug_message ( "\n" ) ;
    } 

  buffer_pointer = buffer ;

  unsigned char BCC = TLP_ACK ;
  put_socket_byte ( TLP_ACK )  ;

  BCC ^= (unsigned char)(length+1) ;
  put_socket_byte ( (unsigned char)(length+1) )  ;

  BCC ^= OSB ;
  put_socket_byte ( OSB )  ;

  for ( i = 0 ; i < length ; i ++ )
    {
      BCC ^= output_buffer [ i ] ;
      put_socket_byte ( output_buffer [ i ] )  ;
    }

  reset_output_buffer ( ) ;

  put_socket_byte ( BCC ) ;

  return send_buffer ( connection_socket ) ;
}



bool TLP224::get_debug_TLP224_traffic ( )
{
  return debug_TLP224_traffic ;
}

bool TLP224::set_debug_TLP224_traffic ( bool b )
{
  return debug_TLP224_traffic = b ;
}


bool TLP224::put_byte_into_output_buffer ( unsigned char b )
{
   if ( output_buffer_end <= output_buffer_pointer )
      return false ;
   * ( output_buffer_pointer ++ ) = b ;
   return true ;
}


bool TLP224::get_byte_from_input_buffer ( unsigned char & b )
{
   if ( input_buffer_end <= input_buffer_pointer )
      return false ;
   b = * ( input_buffer_pointer ++ ) ;
   return true ;
}

void TLP224::reset_output_buffer ( )
{
  output_buffer_pointer = output_buffer;
}


bool TLP224::initialize_socket_io ( SOCKET & socket_card_socket )
{
  unsigned short int wVersionRequested;
  WSADATA wsaData;
  int err;
 
  wVersionRequested = MAKEWORD( 2 , 0 ); /* We need 2.0 functionality */
 
  err = WSAStartup( wVersionRequested, &wsaData );
  if ( err != 0 ) 
    {
      error_message ( "Couldn't find Winsock DLL in initialize_socket_io!\n");
      _flushall();
      return false;
    }
 
  /* Confirm that the WinSock DLL supports 1.1.*/
  /* Note that if the DLL supports versions greater    */
  /* than 1.1 in addition to 1.1, it will still return */
  /* 1.1 in wVersion since that is the version we      */
  /* requested.                                        */
 
  if ( LOBYTE( wsaData.wVersion ) != 2
      || HIBYTE( wsaData.wVersion ) != 0 )
    {
      error_message ( "WinSock 1.1 not supported in initialize_socket_io!\n");
      _flushall();
      WSACleanup( );
      return false;
    }
 
  /* The WinSock DLL is acceptable. */  

  if ( INVALID_SOCKET == ( socket_card_socket = socket ( AF_INET , SOCK_STREAM , IPPROTO_IP ) ) ) 
    {
      error_message ( "Couldn't create socket initialize_socket_io!\n");
      _flushall();
      return false;
    }

  return true;
}


bool TLP224::initialize_socket_server ( SOCKET & socket_card_socket ) 
{
  struct sockaddr_in
    {
      short   sin_family;
      u_short sin_port;
      u_long sin_addr;
      char    sin_zero[8];
    } name =
    {
      AF_INET ,
      htons(9025) ,
      htonl(INADDR_ANY) ,
      0, 0, 0, 0, 0, 0, 0, 0
    } ;

  if ( 0 != bind ( socket_card_socket , ( LPSOCKADDR ) &name, sizeof ( name ) ) )
    {
      error_message ( "Couldn't bind socket to port in initialize_socket_io!\n");
      return false;
    }

  if ( SOCKET_ERROR == listen ( socket_card_socket , 5 ) )
    {
      error_message ( "Couldn't listen to socket in initialize_socket_io!\n");
      return false;
    }

    struct sockaddr_in cli_addr;
    int clilen = sizeof(cli_addr);

  if ( INVALID_SOCKET == 
//       ( socket_card_connection_socket =
       ( socket_card_socket =
           accept ( socket_card_socket , (struct sockaddr *) &cli_addr, &clilen) ) )
    {
      error_message ( "Couldn't accept connections to socket in initialize_socket_io!\n");
      return false;
    }

  return true;
}


bool TLP224::initialize_socket_client ( SOCKET & socket_card_socket, const char * host, u_short port , bool report_error , bool report_success) 
{
  int result = NO_ERROR ;
  struct addrinfo * AddressInformationPtr;

  struct sockaddr_in
    {
      short   sin_family;
      u_short sin_port;
      u_long  sin_addr;
      char    sin_zero[8];
    };

  sockaddr_in name =
    {
      AF_INET ,
      htons ( port ) , // tlp port is 9025
      0x0100007f , // 127.0.0.1 (localhost)
      // 0x7e00000a , // 10.0.0.126
      // 0x194626d8 , // 216.38.70.25
      0, 0, 0, 0, 0, 0, 0, 0
    } ;
  
  if ( NULL != host )
  {
    result = getaddrinfo ( host, NULL, NULL, & AddressInformationPtr );
    
    if ( NO_ERROR != result )
    {
      if ( report_error ) 
        error_message ( "Couldn't resolve host \"%s\" in initialize_socket_client!\n" , host );
      return false;
    }

    name . sin_addr = 
      ((struct sockaddr_in *)(AddressInformationPtr ->ai_addr)) -> sin_addr ;
    freeaddrinfo(AddressInformationPtr);
  }
    
  if ( SOCKET_ERROR == connect ( socket_card_socket, ( LPSOCKADDR ) &name, sizeof(name) ) )
    {
      if ( report_error ) 
        error_message ( "Couldn't connect to socket in initialize_socket_client!\n");
      return false;
    }
  else
    {
      if ( report_success )
        debug_message ( "initialize_socket_client: connect(%04X) succeeded.\n",
                        socket_card_socket ) ;
    }

  return true;
}


void TLP224::finalize_socket_io ( SOCKET & socket_card_socket) 
{
  if ( INVALID_SOCKET == socket_card_socket )
    return ;

  shutdown ( socket_card_socket , SD_SEND ) ;

  if ( SOCKET_ERROR == closesocket ( socket_card_socket ) )
    {
      error_message ( "Couldn't close socket in finalize_socket_io!\n" ) ;
      return ;
    }
}
