// Virtual Smart Card Reader
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include "stdafx.h"
#include "Debug.h"
#include "TLP_Server.h"

#define MAXIMUM_TLP_DATA_BYTES 250

const char TLP_Server::default_hostname[] = "localhost";

long TLP_Server::initialize( )
{
	m_bIntError = true;

	if (m_hostname == NULL) 
	{
		m_hostname = default_hostname ;
		m_port = DEFAULT_APDU_PORT;
	}
 
	return ! (m_ServerTLP224IO.initialize_socket_io(m_scktCardEdge) &&
			  m_ServerTLP224IO.initialize_socket_client(m_scktCardEdge,
											            (const char *) m_hostname,
											            (u_short) m_port,
											            m_bIntError,
												        true));
}

long TLP_Server::finalize( )
{
	m_ServerTLP224IO.finalize_socket_io(m_scktCardEdge);
  
	return 0;
}

long TLP_Server::power_up(void)
{
    unsigned char *response = NULL;
	int num_received;

	static const unsigned char SHORT_TLP_RESET_COMMAND[] = 
    { /*TLP_ACK*/ /*TLP_LEN=0x04*/ /*OSB= 0x6E*/ 0x00, 0x00, 0x00 /* BCC=0x0A */ /* ETX */ };

	for (int i = 0; i < sizeof(SHORT_TLP_RESET_COMMAND); i++)
	{
		m_ServerTLP224IO.put_byte_into_output_buffer(SHORT_TLP_RESET_COMMAND[i]);
	}

	if (! m_ServerTLP224IO.send_TLP224_packet(m_scktCardEdge, 0x6E, sizeof(SHORT_TLP_RESET_COMMAND)))
    {
		error_message("Connection broken during power up!\n");
		return 1 ;
    }

	unsigned char tlp_command;
  
	if (! m_ServerTLP224IO.receive_TLP224_packet(m_scktCardEdge, tlp_command, num_received))
    {
		error_message("Connection broken during power up!\n");
		return 1;
    }

    response = new unsigned char[num_received];

	for (int i = 0; i < num_received; i++)
	{
		m_ServerTLP224IO.get_byte_from_input_buffer(response[i]);
	}

	if (! (0x00 == tlp_command && 0x28 == response[0] && 0x02 == response[1] &&
           num_received - 3 == response[2]))
    {
		error_message("Unexpected response from power up!\n");
		return 1;
    }

	m_ATRlength = num_received - 3;
	memcpy(m_ATR, &response[3], m_ATRlength);
    
    delete [] response;

	return 0;
}

long TLP_Server::power_down(void) 
{
	m_ServerTLP224IO.send_TLP224_packet(m_scktCardEdge, 0x4D, 0x00);  // 0x4D == TLP POWER DOWN COMMAND

	return 0;
}

void TLP_Server::get_ATR(unsigned char *ATR, unsigned int &ATR_length )
{
    ATR_length = m_ATRlength;
    memcpy(ATR, m_ATR, m_ATRlength); 
}

int TLP_Server::set_hostname(const char *passed_hostname)
{
	if (strlen(passed_hostname) < MAX_HOSTNAME_BUFFER_LENGTH)  // check if the length is reasonable
	{
		m_hostname = passed_hostname; 
		return 0;
	}

	return 1;  // Didn't know how to handle it.
}

bool TLP_Server::exchangeData(BYTE *dataSend, int dataSendSize, BYTE *dataRecv, int &dataRecvSize)
{
	BYTE tlpCommand;
	int numBytesRecv = 0;

	if(dataSendSize <= MAXIMUM_TLP_DATA_BYTES) {
		bool bSuccess;

		for(int i = 0; i < dataSendSize; i++)
		{
			m_ServerTLP224IO.put_byte_into_output_buffer(dataSend[i]);
		}

		// send out TLP224 data packet
		if ( dataSendSize - 5 < dataSend[4] ) {

			bSuccess = m_ServerTLP224IO.send_TLP224_packet(m_scktCardEdge, 
							TLP224::TLPC_INPUT_FROM_MICROPROCESSOR, dataSendSize);
		}
		else {

			bSuccess = m_ServerTLP224IO.send_TLP224_packet(m_scktCardEdge, 
							TLP224::TLPC_OUTPUT_TO_MICROPROCESSOR, dataSendSize);
		}

		if (!bSuccess) {

			error_message("Connection broken while sending data!\n");
			dataRecvSize = 0;
			return false;
		}
	}
	else //send in 248 byte packets
	{
		int totalBytesSent = 0;

		while((dataSendSize-totalBytesSent) > MAXIMUM_TLP_DATA_BYTES)
		{
			for(unsigned int i = 0; i < MAXIMUM_TLP_DATA_BYTES; i++)
			{
				m_ServerTLP224IO.put_byte_into_output_buffer(dataSend[totalBytesSent+i]);
			}
			if(! m_ServerTLP224IO.send_TLP224_packet(m_scktCardEdge,  //send out TLP224 data packet
													 TLP224::TLPC_OUTPUT_TO_MICROPROCESSOR,
	 												 MAXIMUM_TLP_DATA_BYTES))
			{
				error_message("Connection broken while sending data!\n");
				dataSendSize = 0;
				dataSend = NULL;
				return false;
			}
			totalBytesSent += MAXIMUM_TLP_DATA_BYTES;
		}
		
		for(int i = 0; i < dataSendSize-totalBytesSent; i++)  //send last packet, less than 248 bytes
		{
			m_ServerTLP224IO.put_byte_into_output_buffer(dataSend[totalBytesSent+i]);
		}
		if(! m_ServerTLP224IO.send_TLP224_packet(m_scktCardEdge,  //send out TLP224 data packet
												 TLP224::TLPC_OUTPUT_TO_MICROPROCESSOR,
	 											 dataSendSize-totalBytesSent))
		{
			error_message("Connection broken while sending data!\n");
			dataSendSize = 0;
			dataSend = NULL;
			return false;
		}
	}

	if(! m_ServerTLP224IO.receive_TLP224_packet(m_scktCardEdge, tlpCommand, numBytesRecv))  //receive first packet
	{
		error_message("Connection broken while receiving data!\n");
		dataRecvSize = 0;
		dataRecv = NULL;
		return false;
    }

#ifdef _DEBUG
	char szDebug[128];

	sprintf(szDebug, "numBytesRecv = %d\n", numBytesRecv);
	PIVVSCR_DEBUG(szDebug);
#endif

	for(int i = 0; i < numBytesRecv; i++)
	{
		m_ServerTLP224IO.get_byte_from_input_buffer(dataRecv[i]);
	}

	if(numBytesRecv <= MAXIMUM_TLP_DATA_BYTES)  //just one packet coming back
	{
		dataRecvSize = numBytesRecv;
		return true;
	}
	else  //two or more packets coming back
	{
		int sizeOfDataIn = MAXIMUM_TLP_DATA_BYTES * 2;

		while(numBytesRecv == MAXIMUM_TLP_DATA_BYTES)
		{
			if(! m_ServerTLP224IO.receive_TLP224_packet(m_scktCardEdge, tlpCommand, numBytesRecv))
			{
				error_message("Connection broken while receiving data!\n");
				dataRecvSize = 0;
				return false;
			}
			if(numBytesRecv == 0) //special case where we receive one packet of 248 bytes exactly
			{
				dataRecvSize = MAXIMUM_TLP_DATA_BYTES;
				return true;
			}
			if(numBytesRecv < MAXIMUM_TLP_DATA_BYTES)  //last packet, less than 248 bytes
			{
				BYTE *saveBuffer = new BYTE[sizeOfDataIn+numBytesRecv];  //make room for new data coming back
				memcpy(saveBuffer, dataRecv, sizeOfDataIn);
				dataRecv = saveBuffer;

				for(i = 0; i < numBytesRecv; i++)
				{
					m_ServerTLP224IO.get_byte_from_input_buffer(dataRecv[sizeOfDataIn-
																         MAXIMUM_TLP_DATA_BYTES+
																         i]);
				}
				dataRecvSize = sizeOfDataIn + numBytesRecv - MAXIMUM_TLP_DATA_BYTES;
				return true;
			}

			BYTE *saveBuffer = new BYTE[sizeOfDataIn];  //make room for new data coming back
			memcpy(saveBuffer, dataRecv, sizeOfDataIn);
			dataRecv = saveBuffer;

			for(i = 0; i < numBytesRecv; i++)
			{
				m_ServerTLP224IO.get_byte_from_input_buffer(dataRecv[sizeOfDataIn-MAXIMUM_TLP_DATA_BYTES+i]);
			}
			sizeOfDataIn += numBytesRecv;
		}
		dataRecvSize = sizeOfDataIn - 2*MAXIMUM_TLP_DATA_BYTES;
		return true;
	}

	return false;
}