#include "stdAfx.h"
#include "winscard.h"
#include "WDMIoctl.h"
#include "PIV VSCR Service.h"
#include "Debug.h"
#include "VirtualSmartCardReader.h"
#include "TLP_Server.h"
#include "Registry.h"

		// A. offset 0: set 55h by this app as if the card is inserted.
#define EBDAOffset_55flag				0x00
		// B. offset 1: data request flag (set by the Virtual Card Driver)
#define EBDAOffset_DataRequestflag		0x01
		// C. offset 2: data ready flag (set by this appliacation if the return data is ready)
#define EBDAOffset_DataReadyflag		0x02
        // D. offset 3: reserved
		// E. offset 4-7: CTRL code from the resources manager or the card application
#define EBDAOffset_CTLCode				0x04
		// F. offset 8-9: CBIN
#define EBDAOffset_CBIN					0x08
		// G. offset 10--: IN Data
#define EBDAOffset_INData				0x0A
		// H. after IN Data: CBOUT
#define EBDAOffset_CBOUT				0x0B
		// I. after CBOUT: the Data to be return
#define EBDAOffset_OUTData				0x0D

// error status word when application is not connected
#define NOT_CONNECTED_SW 0x6F00 

TLP_Server g_tlpServer;
ULONG g_ulCardState = SCARD_ABSENT;
ULONG g_ulCurrentProtocol = 0;

CVirtualSmartCardReader::CVirtualSmartCardReader(void)
{
	m_hDevice = INVALID_HANDLE_VALUE;
	m_hostname = _T("");
	m_port = 0;
	m_bAutoInsert = FALSE;
	m_iPacketCnt = 0;
	m_bConnectedtoRefImp = FALSE;
	m_bWDMFlag = FALSE;
}

CVirtualSmartCardReader::~CVirtualSmartCardReader(void)
{
	if (m_hDevice != INVALID_HANDLE_VALUE) {

		CloseHandle(m_hDevice);
		m_hDevice = INVALID_HANDLE_VALUE;
	}
}

BOOL CVirtualSmartCardReader::GetConfigurationFromRegistry()
{
	DWORD dwSize = 0;
	BOOL bSuccess = FALSE;

	// get size of host name registry value
	if (GetRegistryValueSize(
				HKEY_LOCAL_MACHINE, 
				PIVVSCR_PARAMETERS_REGKEY, 
				PIVVSCR_HOSTNAME_REGVALUE, 
				&dwSize) == ERROR_SUCCESS) {
		CString csStrHostName(' ', dwSize); // allocate buffer

		// get host name
		if (GetRegistryValue(
					HKEY_LOCAL_MACHINE, 
					PIVVSCR_PARAMETERS_REGKEY, 
					PIVVSCR_HOSTNAME_REGVALUE, 
					&csStrHostName, 
					dwSize) == ERROR_SUCCESS) {

			m_hostname = csStrHostName;

			// get port number
			if (GetRegistryValue(
						HKEY_LOCAL_MACHINE, 
						PIVVSCR_PARAMETERS_REGKEY, 
						PIVVSCR_PORT_REGVALUE, 
						&m_port) == ERROR_SUCCESS) {
				DWORD dwAutoInsert;

				// get auto insert value
				if (GetRegistryValue(
							HKEY_LOCAL_MACHINE, 
							PIVVSCR_PARAMETERS_REGKEY, 
							PIVVSCR_AUTOINSERT_REGVALUE, 
							&dwAutoInsert) != ERROR_SUCCESS) {

					m_port = 0;
					m_hostname = _T("");

					// log error in event log
					ReportPIVEvent(
							PIV_EVENT_ERROR, 
							"Could not get value of registry key %s at path \\\\HKEY_LOCAL_MACHINE\\%s\n", 
							PIVVSCR_AUTOINSERT_REGVALUE, 
							PIVVSCR_PARAMETERS_REGKEY);
				}
				else {

					m_bAutoInsert = (dwAutoInsert > 0) ? TRUE : FALSE;
					bSuccess = TRUE;
				}
			}
			else {

				m_hostname = _T("");

				// log error in event log
				ReportPIVEvent(
						PIV_EVENT_ERROR, 
						"Could not get value of registry key %s at path \\\\HKEY_LOCAL_MACHINE\\%s\n", 
						PIVVSCR_PORT_REGVALUE, 
						PIVVSCR_PARAMETERS_REGKEY);
			}
		}
		else {

			// log error in event log
			ReportPIVEvent(
					PIV_EVENT_ERROR,
					"Could not get value of registry key %s at path \\\\HKEY_LOCAL_MACHINE\\%s\n", 
					PIVVSCR_HOSTNAME_REGVALUE, 
					PIVVSCR_PARAMETERS_REGKEY);
		}
	}
	else {

		// log error in event log
		ReportPIVEvent(
				PIV_EVENT_ERROR,
				"Could not get size of registry key %s at path \\\\HKEY_LOCAL_MACHINE\\%s.", 
				PIVVSCR_HOSTNAME_REGVALUE, 
				PIVVSCR_PARAMETERS_REGKEY);
	}

	return bSuccess;
}

BOOL CVirtualSmartCardReader::isConnected()
{
	return m_bConnectedtoRefImp;
}

BOOL CVirtualSmartCardReader::Initialize()
{
	BOOL bSuccess = FALSE;

	// make sure we don't already have a handle to the memory access driver
	if (m_hDevice != INVALID_HANDLE_VALUE)
		return FALSE;

    m_iPacketCnt = 0;

	// get handle to memory access driver
    m_hDevice = CreateFile (
						PIV_MEMDRV_DEVICE_NAME,
						GENERIC_READ | GENERIC_WRITE,
						FILE_SHARE_READ,
						NULL,
						OPEN_EXISTING,
						FILE_ATTRIBUTE_NORMAL,
						NULL);

	if (m_hDevice != INVALID_HANDLE_VALUE) {

		// get configuration info from registry
		if (GetConfigurationFromRegistry()) {

			m_bWDMFlag = TRUE;

			// Set the flag to indicate that this application is running
			// to the Virtual Smart Card Driver
			WriteDW_EBDA(EBDAOffset_55flag, 0x55, 0x01);			// Write 55h to EBDA offset 0
			WriteDW_EBDA(EBDAOffset_DataRequestflag, 0x00, 0x01);	// Write 00h to EBDA offset 1
			WriteDW_EBDA(EBDAOffset_DataReadyflag, 0x00, 0x01);		// Write 00h to EBDA offset 2

			bSuccess = TRUE;
		}
		else {

			m_bWDMFlag = FALSE;

			CloseHandle(m_hDevice);
			m_hDevice = INVALID_HANDLE_VALUE;

			// log error in event log
			ReportPIVEvent(PIV_EVENT_ERROR, "Failed to retrieve configuration information from the registry.");
		}
	}
	else {

		// log error in event log
		ReportPIVEvent(PIV_EVENT_ERROR, "Memory access driver not installed.");

		m_bWDMFlag = FALSE;
	}

	return bSuccess;
}

BOOL CVirtualSmartCardReader::Connect()
{
	BOOL bSuccess = FALSE;

	// set TLP Server host parameter
	g_tlpServer.set_hostname((const char *)m_hostname.GetBuffer());
	m_hostname.ReleaseBuffer();

	// set TLP Server host port
	g_tlpServer.set_port(m_port);

	if (0 != g_tlpServer.initialize() || 0 !=  g_tlpServer.power_up()) {

		m_bConnectedtoRefImp = FALSE;

		// log error in event log
		ReportPIVEvent(PIV_EVENT_ERROR, "Failed to connect to the TLP server.");
	}
	else {
		BYTE *ATR = new BYTE[32];
		CString csATR, cByte;
		unsigned int ATRlen = 0;

		m_bConnectedtoRefImp = TRUE;

		g_tlpServer.get_ATR(ATR, ATRlen);

		for (unsigned int i = 0; i < ATRlen; i++) {

			cByte.Format("%02X ", ATR[i]);
			csATR += cByte;
		}

		// log success in event log
		ReportPIVEvent(PIV_EVENT_SUCCESS, "Connected - %s", csATR);

		delete [] ATR;
		ATR = NULL;

		bSuccess = TRUE;
	}

	return bSuccess;
}

void CVirtualSmartCardReader::Disconnect()
{
	m_bConnectedtoRefImp = FALSE;

	// power down TLP Server
    g_tlpServer.power_down();
    g_tlpServer.finalize();
}

void CVirtualSmartCardReader::Terminate()
{
	// Clear the flag to indicate that this application is not running
	// to the Virtual Smart Card Driver
    WriteDW_EBDA(EBDAOffset_55flag, 0x00, 0x01);			// Write 00h to EBDA offset 0
    WriteDW_EBDA(EBDAOffset_DataRequestflag, 0x00, 0x01);	// Write 00h to EBDA offset 1
    WriteDW_EBDA(EBDAOffset_DataReadyflag, 0x00, 0x01);		// Write 00h to EBDA offset 2

	// close handle to memory access driver
	if (m_hDevice != INVALID_HANDLE_VALUE) {

		CloseHandle(m_hDevice);
		m_hDevice = INVALID_HANDLE_VALUE;
	}

	if (m_bConnectedtoRefImp) {

		m_bConnectedtoRefImp = FALSE;

		// power down TLP Server
		g_tlpServer.power_down();
		g_tlpServer.finalize();
	}
}

void CVirtualSmartCardReader::InsertCard()
{
	HANDLE hDevice;

	// get handle to VSCR driver
    hDevice = CreateFile (
					PIV_VSCRDRV_DEVICE_NAME,
					GENERIC_ALL,
					FILE_SHARE_READ,
					NULL,
					OPEN_EXISTING,
					0,
					NULL);

	if (hDevice != INVALID_HANDLE_VALUE) {
		DWORD dwBytesReturned;

		// send command to remove card from reader
		DeviceIoControl(
				hDevice,
				IOCTL_SMARTCARD_INSERT,
				NULL,
				0,
				NULL,
				0,
				&dwBytesReturned,
				NULL);

		CloseHandle(hDevice);
	}
}

void CVirtualSmartCardReader::ForceCardRemoval()
{
	HANDLE hDevice;

	// get handle to VSCR driver
    hDevice = CreateFile (
					PIV_VSCRDRV_DEVICE_NAME,
					GENERIC_ALL,
					FILE_SHARE_READ,
					NULL,
					OPEN_EXISTING,
					0,
					NULL);

	if (hDevice != INVALID_HANDLE_VALUE) {
		DWORD dwBytesReturned;

		// send command to remove card from reader
		DeviceIoControl(
				hDevice,
				IOCTL_SMARTCARD_REMOVE,
				NULL,
				0,
				NULL,
				0,
				&dwBytesReturned,
				NULL);

		CloseHandle(hDevice);
	}
}

void CVirtualSmartCardReader::Process()
{
	SHORT nOffsetEBDA;
	ULONG nDataDW;
	ULONG status;
	ULONG nReadData;
	ULONG nWriteData;
	ULONG nReturnCount;
	ULONG ctlcode;
	SHORT nInCount;
    CString str;
	SHORT i;

	UCHAR revbuffer[300];
	UCHAR sendAPDUbuffer[300];
	UCHAR recAPDUbuffer[300];
	int sendsize, recsize;

	if (m_bWDMFlag) {

        // A. offset 0: set 55h by this app as if the card is inserted.
		// B. offset 1: data request flag (set by the Virtual Card Driver)
		// C. offset 2: data ready flag (set by this appliacation if the return data is ready)
        // D. offset 3: reserved
		// E. offset 4-7: CTRL code from the resources manager or the card application
		// F. offset 8-9: CBIN
		// G. offset 10--: IN Data
		// H. after IN Data: CBOUT
		// I. after CBOUT: the Data to be return

		//
		// 1. Check the flag if the Data is requested (Offset 1)
		//
		nDataDW = ReadDW_EBDA(EBDAOffset_DataRequestflag, 1);	// Read byte data from EBDA

        if (nDataDW == 1) {

			//
            // 2. Clear the Data Request Flag
			//
			WriteDW_EBDA(EBDAOffset_DataRequestflag, 0x00, 0x01);	// Clear the Data Requeset Flag

            //
			// 3. Read the CTRL code from EBDA
			//
            ctlcode = ReadDW_EBDA(EBDAOffset_CTLCode, 4);	// Read data from EBDA
/*
#ifdef TRAFFIC
            // ** Add a row to the list control (column 0)
            str.Format("0x%lX", ctlcode);
            SHORT nItem = m_lcCode.InsertItem(m_iPacketCnt++, str);
#endif
*/

            //
			// 4. Read the CBIN (Count Byte for IN) from EBDA
			//
            nDataDW = ReadDW_EBDA(EBDAOffset_CBIN, 2);	// Read data from EBDA
            nInCount = (SHORT)nDataDW;
/*
#ifdef TRAFFIC
            // ** Add a cell to the list control (column 1)
			str.Format("0x%X", (UCHAR)nDataDW);
            m_lcCode.SetItemText(nItem, 1, str);
#endif
*/
            //
			// 5. Read the IN Data from EBDA
			//
			str = "";
			if (nInCount > 0) {

				nReadData = 0;

				for (i = 0; i < nInCount; i++) {
                    nDataDW = ReadDW_EBDA((EBDAOffset_CBIN + 1 + nInCount - i), 1);	// Read data from EBDA
					//formatting string
					if (nDataDW < 0x0F)
						str.AppendFormat("%lX", 0);
					str.AppendFormat("%lX", nDataDW);
					nReadData = nReadData * 256 + nDataDW;

					// for transmit IOCTL, the IN data can be larger than a 
					// DWORD so we need to store it in a buffer
					if (ctlcode == IOCTL_SMARTCARD_TRANSMIT)
					    revbuffer[i] = (UCHAR)nDataDW;
				}
/*
#ifdef TRAFFIC
                // ** Add a cell to the list control (column 2)
                m_lcCode.SetItemText(nItem, 2, str);
#endif
*/
			}

            //
			// 6. Read the OUT Count to be returned
			//
			nDataDW = ReadDW_EBDA((EBDAOffset_INData + nInCount), 2);	// Read data from EBDA
/*
#ifdef TRAFFIC
            // ** Add a row to the list control (column 3)
            str.Format("0x%X", nDataDW);
            m_lcCode.SetItemText(nItem, 3, str);
#endif
*/
            //
			// 7. Read the OUT Count to be return
			//
			if (ctlcode == IOCTL_SMARTCARD_GET_STATE) {

				// write the data count to return
				nReturnCount = 4;
				WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

				nWriteData = g_ulCardState;
				str.Format("0x%X", nWriteData);
				WriteDW_EBDA((EBDAOffset_OUTData + nInCount), nWriteData, 0x04);	// Write ULONG

				// write the status to return to the Smart Card Driver
				status = SCARD_S_SUCCESS;
			}
			else if (ctlcode == IOCTL_SMARTCARD_IS_PRESENT) {

				if (g_ulCardState > SCARD_ABSENT) {
					str = "(is_present: true)";
					// write the status to return to the Smart Card Driver
					status = SCARD_S_SUCCESS;
				} 
				else {
					str = "(is_present: false)";
					status = 0x00000103;		// STATUS_PENDING
				}
				// No data to return
				// write the data count to return
				nReturnCount = 0x00;
				WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);
			}
			else if (ctlcode == IOCTL_SMARTCARD_POWER) {
				if (nReadData == SCARD_POWER_DOWN) {//RM is telling Reader to power down
					str = "(powerdown)";
					g_ulCardState = SCARD_PRESENT;
					g_ulCurrentProtocol = 0;

					// write the data count to return
					nReturnCount = 0x00;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);
				}
				else {//RM is telling Reader to power up/reset

					// No data to return
					str = "(powerup/reset)"; //SCARD_COLD_RESET
					g_ulCardState = SCARD_NEGOTIABLE;

					// write the data count to return
					nReturnCount = 5;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);
					// write the data to return ATR
					nWriteData = 0x4096903B; //
					nOffsetEBDA = EBDAOffset_OUTData + nInCount;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					nWriteData = 0x0000000A; //
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 4;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					str = "ATR";
				}

				// write the status to return to the Smart Card Driver
				status = SCARD_S_SUCCESS;
            }
			else if (ctlcode == IOCTL_SMARTCARD_IS_ABSENT) {

				if (g_ulCardState > SCARD_ABSENT) {
					str = "(is_absent: false)";
					status = 0x00000103;	// STATUS_PENDING
				} 
				else {
					str = "(is_absent: true)";
					// write the status to return to the Smart Card Driver
					status = SCARD_S_SUCCESS;	// status = SCARD_S_SUCCESS
				}
				// No data to return
				// write the data count to return
				nReturnCount = 0x00;
				WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);
			}
			else if (ctlcode == IOCTL_SMARTCARD_SET_ATTRIBUTE) {

				// No data to return
				str = "(set_attribute)";
				// write the data count to return
				nReturnCount = 0x00;
				WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

				// write the status to return to the Smart Card Driver
				status = 0xC00000BB;			// status = STATUS_NOT_SUPPORTED
			}
			else if (ctlcode == IOCTL_SMARTCARD_TRANSMIT) {

				PIVVSCR_DEBUG("------- IOCTL_SMARTCARD_TRANSMIT -------\n");

				if (m_bConnectedtoRefImp) {
					UINT j;

					sendsize = nInCount - 8;
					// convert to APDU
					for (i=0; i<sendsize; i++){
						sendAPDUbuffer[i] = revbuffer[nInCount-i-1-8];
					}

#ifdef _DEBUG
					char szDebug[1280], szDebugData[1024];

					SvcDebugHexToStr(sendAPDUbuffer, sendsize, szDebugData, sizeof(szDebugData));
					sprintf(szDebug, "send APDU = 0x%s\n", szDebugData);
					PIVVSCR_DEBUG(szDebug);
#endif

					PIVVSCR_DEBUG("Calling exchangeData...\n");

					g_tlpServer.exchangeData(sendAPDUbuffer, sendsize, recAPDUbuffer, recsize);

					PIVVSCR_DEBUG("Finished exchanging data\n");

#ifdef _DEBUG
					SvcDebugHexToStr(recAPDUbuffer, recsize, szDebugData, sizeof(szDebugData));
					sprintf(szDebug, "rcv APDU = 0x%s\n", szDebugData);
					PIVVSCR_DEBUG(szDebug);
#endif

					str.FormatMessage(_T("%1!d! Bytes Received"), recsize);

					// write the data count to return
					nReturnCount = 8 + recsize;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

					// write header
					nWriteData = 0x00000001; //
					nOffsetEBDA = EBDAOffset_OUTData + nInCount;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					nWriteData = 0x00000008; //
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 4;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);

					// write return data back into EBDA in blocks of 4 bytes
					nWriteData = 0;
					for(i = recsize; i >= 4 ; i -= 4) {
						for (j = 4; j > 0; j--) {
							nWriteData = nWriteData * 256 + recAPDUbuffer[(recsize-i)+(j-1)];
						}
						nOffsetEBDA = EBDAOffset_OUTData + nInCount + 8 + (recsize-i);
						WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					}
					nWriteData = 0;
					while ( i > 0) {
						nWriteData = nWriteData * 256 + recAPDUbuffer[((recsize/4)*4)+(i-1)];
						i--;
					}
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 8 + ((recsize/4)*4) ;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
				}
				else {	//RefImp not connected
					str.FormatMessage(_T("Not connected: sending error SW"));
					nReturnCount = 10;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);
					// write header
					nWriteData = 0x00000001; //
					nOffsetEBDA = EBDAOffset_OUTData + nInCount;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					nWriteData = 0x00000008; //
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 4;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					// write return data back into EBDA in blocks of 4 bytes
					nWriteData = ((NOT_CONNECTED_SW & 0x000000FF) << 8) + ((NOT_CONNECTED_SW & 0x0000FF00) >> 8);
					//nWriteData = (nWriteData * 256) * 256;
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 8;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
				}
				// write the status to return to the Smart Card Driver
				status = SCARD_S_SUCCESS;
            }
			else if ( ctlcode == IOCTL_SMARTCARD_GET_ATTRIBUTE){

				if ( nReadData == SCARD_ATTR_CHARACTERISTICS ) {

					// write the data count to return
					nReturnCount = 4;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

					// write the data to return
					nWriteData = 0x00000000;
					str.Format("0x%X", nWriteData);
					nOffsetEBDA = EBDAOffset_OUTData + nInCount;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);	// Write ULONG

					// write the status to return to the Smart Card Driver
					status = SCARD_S_SUCCESS;
				}
				else if ( nReadData == SCARD_ATTR_VENDOR_NAME ) {

					// write the data count to return
					nReturnCount = 0;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

					str = "";

					// write the status to return to the Smart Card Driver
					status = SCARD_S_SUCCESS;
				}
				else if ( nReadData == SCARD_ATTR_VENDOR_IFD_TYPE ) {

					// write the data count to return
					nReturnCount = 25;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

					// write the data to return : Virtual Smart Card Reader
					nWriteData = 0x74726956; // triV
					nOffsetEBDA = EBDAOffset_OUTData + nInCount;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					nWriteData = 0x206c6175; // _lau
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 4;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					nWriteData = 0x72616d53; // ramS
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 8;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					nWriteData = 0x61432074; // aC_t
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 12;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					nWriteData = 0x52206472; // R_dr
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 16;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					nWriteData = 0x65646165; // edae
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 20;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					nWriteData = 0x00000072; // ___r
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 24;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);

					str = "Virtual Smart Card Reader";

					// write the status to return to the Smart Card Driver
					status = SCARD_S_SUCCESS;
				}
				else if ( nReadData == SCARD_ATTR_ATR_STRING ) {

					// write the data count to return
					nReturnCount = 5;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);
					// write the data to return ATR
					nWriteData = 0x4096903B; //
					nOffsetEBDA = EBDAOffset_OUTData + nInCount;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					nWriteData = 0x0000000A; //
					nOffsetEBDA = EBDAOffset_OUTData + nInCount + 4;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);
					str = "ATR";
					// write the status to return to the Smart Card Driver
					status = SCARD_S_SUCCESS;
				}
				else if ( nReadData == SCARD_ATTR_CURRENT_PROTOCOL_TYPE ) {

					// write the data count to return
					nReturnCount = 4;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

					nWriteData = g_ulCurrentProtocol; //
					nOffsetEBDA = EBDAOffset_OUTData + nInCount;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);

					str = "Protocol";
					// write the status to return to the Smart Card Driver
					status = SCARD_S_SUCCESS;
				}
				else if (nReadData == SCARD_ATTR_DEVICE_UNIT) {

					// write the data count to return
					nReturnCount = 4;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

					// write the data to return
					nWriteData = 0x00000000;
					str.Format("0x%X", nWriteData);
					nOffsetEBDA = EBDAOffset_OUTData + nInCount;
					WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);

					// write the status to return to the Smart Card Driver
					status = SCARD_S_SUCCESS;
				}
				else {

					// TODO: Do not return any data (temp)
					str = " Not supported ";
					// write the data count to return
					nReturnCount = 0x00;
					WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

					// write the status to return to the Smart Card Driver
					status = 0xC00000BB;	// status = STATUS_NOT_SUPPORTED
				}
            }
			else if ( ctlcode == IOCTL_SMARTCARD_SET_PROTOCOL){

				g_ulCardState = SCARD_SPECIFIC;
				g_ulCurrentProtocol = 1;		// set to T=0

				// write the data count to return
				nReturnCount = 4;
				WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

				nWriteData = g_ulCurrentProtocol;
				nOffsetEBDA = EBDAOffset_OUTData + nInCount;
				WriteDW_EBDA(nOffsetEBDA, nWriteData, 0x04);

				str = "Protocol";
				// write the status to return to the Smart Card Driver
				status = SCARD_S_SUCCESS;
			}
			else if (ctlcode == IOCTL_SMARTCARD_INSERT) {

				if (m_bConnectedtoRefImp) {

					str = "(already inserted)";
					g_ulCardState = SCARD_PRESENT;
					status = SCARD_S_SUCCESS;
				} 
				else {
					str = "(insert)";

					// connect to the card simulator
					if (Connect()) {

						g_ulCardState = SCARD_PRESENT;
						status = SCARD_S_SUCCESS;	// status = SCARD_S_SUCCESS
					} else {

						g_ulCardState = SCARD_ABSENT;
						status = 0xC0000321;		// status = STATUS_SMARTCARD_SUBSYSTEM_FAILURE
					}
				}

				// add event to event viewer log
				if (status == SCARD_S_SUCCESS)
					ReportDefaultInformationEvent(g_AppName, "Card inserted into reader.");
				else
					ReportDefaultErrorEvent(g_AppName, "Card could not be inserted into reader.  Failed to establish connection with card simulator.");

				// No data to return -- write the data count to return
				nReturnCount = 0x00;
				WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);
			}
			else if (ctlcode == IOCTL_SMARTCARD_REMOVE) {

				if (m_bConnectedtoRefImp) {
					str = "(remove)";

					// disconnect from the card simulator
					Disconnect();
					status = SCARD_S_SUCCESS;
				} else {
					str = "(no card)";
					status = 0xC0000178;			// status = STATUS_NO_MEDIA
				}

				g_ulCardState = SCARD_ABSENT;
				g_ulCurrentProtocol = 0;

				// add event to event viewer log
				if (status == SCARD_S_SUCCESS)
					ReportDefaultInformationEvent(g_AppName, "Card removed from reader.");

				// No data to return -- write the data count to return
				nReturnCount = 0x00;
				WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);
			}
			else if (ctlcode == IOCTL_SMARTCARD_EJECT) {
				str = "(eject)";

				// No data to return -- write the data count to return
				nReturnCount = 0x00;
				WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

				//
				// Note: IOCTL_SMARTCARD_EJECT is sent to physically eject a
				// smart card from readers that have an ejection mechanism.
				// When we receive this command, we'll report that card
				// ejection is not supported and instead use our custom 
				// IOCTL_SMARTCARD_REMOVE command to "remove" a card.
				//

				// write the status to return to the Smart Card Driver
				status = 0xC00000BB;			// status = STATUS_NOT_SUPPORTED
			}
			else {

				// No data to return
				str = " ??? ";
				// write the data count to return
				nReturnCount = 0x00;
				WriteDW_EBDA((EBDAOffset_CBOUT + nInCount), nReturnCount, 0x02);

				// write the status to return to the Smart Card Driver
				status = 0xC00000BB;	// status = STATUS_NOT_SUPPORTED
			}

            nOffsetEBDA = EBDAOffset_OUTData + nInCount + (SHORT)nReturnCount;
			WriteDW_EBDA(nOffsetEBDA, status, 0x04);
/*
#ifdef TRAFFIC
			m_lcCode.SetItemText(nItem, 4, str);
			str.Format("0x%X", status);
            m_lcCode.SetItemText(nItem, 5, str);
#endif
*/
			//
			// 8. Set the Data Ready Flag to read the data by the smart cards driver
            //
			nOffsetEBDA = 0x02;	// EBDA offset 2
            WriteDW_EBDA(nOffsetEBDA, 0x01, 0x01);	// Set the Data Ready Flag
		}
	}
}

ULONG CVirtualSmartCardReader::ReadDW_EBDA(USHORT offsetEBDA, UCHAR bCount)
{
	BOOL    bResult;
	ULONG	nOutput;

	if (m_bWDMFlag) {

		//
		// Clear BDA:0xF8 DWORD
		//
		m_byDataIn[0] = (UCHAR)(offsetEBDA % 0x100);	// Address : BDA offset
		m_byDataIn[1] = (UCHAR)(offsetEBDA / 0x100);
		m_byDataIn[2] = bCount;		// Read Byte Count
		m_byDataIn[3] = 0;

		bResult = DeviceIoControl(
						m_hDevice,
						BRIGHTCTRLWDM_IOCTL_BDA,
						&m_byDataIn,
						sizeof(UCHAR) * 4,
						&m_byDataOut,
						sizeof(UCHAR) * bCount,
						&nOutput,
						NULL);
	}
	if (bResult) {
    	if (nOutput == 4) {
            return ((m_byDataOut[3]*256 + m_byDataOut[2])*65536 + (m_byDataOut[1]*256 + m_byDataOut[0]));
	    }
	    else if (nOutput == 2) {
    	    return (m_byDataOut[1]*256 + m_byDataOut[0]);
		}
		else if (nOutput == 1) {
	    	return (m_byDataOut[0]);
		}
	}

	return (0);
}

BOOL CVirtualSmartCardReader::WriteDW_EBDA(USHORT offsetBDA, ULONG DataDW, UCHAR bCount)
{
	BOOL    bResult = false;
	ULONG	nOutput;

	if (m_bWDMFlag) {

		//
		// Clear BDA:0xF8 DWORD
		//
		m_byDataIn[0] = (UCHAR)(offsetBDA % 0x100);	// Address : BDA offset
		m_byDataIn[1] = (UCHAR)(offsetBDA / 0x100);
		m_byDataIn[2] = bCount;		// Write Byte Count
		m_byDataIn[3] = 0;
		m_byDataIn[4] = (UCHAR)(DataDW % 0x100);	// Data byte 0
		m_byDataIn[5] = (UCHAR)(DataDW / 0x100);
		m_byDataIn[6] = (UCHAR)(DataDW / 0x10000);
		m_byDataIn[7] = (UCHAR)(DataDW / 0x1000000);

		bResult = DeviceIoControl(
						m_hDevice,
						BRIGHTCTRLWDM_IOCTL_BDA,
						&m_byDataIn,
						sizeof(UCHAR) * (4+ bCount),
						NULL,
						0,
						&nOutput,
						NULL);
	}

	return bResult;
}
